import collections.abc
import json
import json5
import logging
from typing import Optional, Union
from enum import Enum

import numpy as np
import pandas as pd
from pandas import DataFrame

from shimoku.api.utils import convert_input_data_to_db_items
from shimoku.execution_logger import log_error

logger = logging.getLogger(__name__)


def create_normalized_name(name: str) -> str:
    """Having a name create a normalizedName

    Example
    ----------------------
    # "name": "   Test Borrar_grafico    "
    # "normalizedName": "test-borrar-grafico"
    """
    if any([c not in 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789._- &$'
            for c in name.encode('ascii', 'ignore').decode()]):
        log_error(logger,
                  f'You can only use letters, numbers, spaces, "-" and "_" in your name | '
                  f'you introduced {name}',
                  ValueError
                  )
    # remove empty spaces at the beginning and end
    name: str = name.strip()
    # replace "_" for www protocol it is not good
    name = name.replace('_', '-')

    return '-'.join(name.split(' ')).lower()


def clean_menu_path(menu_path: str) -> tuple[str, str]:
    """
    Break the menu path in the apptype or app normalizedName
    and the path normalizedName if any
    :param menu_path: the menu path
    """
    # remove empty spaces
    menu_path: str = menu_path.strip()
    # replace "_" for www protocol it is not good
    menu_path = menu_path.replace('_', '-')

    try:
        assert len(menu_path.split('/')) <= 2  # we allow only one level of path
    except AssertionError:
        raise ValueError(
            f'We only allow one subpath in your request | '
            f'you introduced {menu_path} it should be maximum '
            f'{"/".join(menu_path.split("/")[:1])}'
        )

    # Split AppType or App Normalized Name
    normalized_name: str = menu_path.split('/')[0]
    name: str = (
        ' '.join(normalized_name.split('-'))
    )

    try:
        path_normalized_name: str = menu_path.split('/')[1]
        path_name: str = (
            ' '.join(path_normalized_name.split('-'))
        )
    except IndexError:
        path_name = None

    return name, path_name


def calculate_percentages_from_list(numbers, round_digits_min):
    """Calculate the proportion of each number in a list
    :param numbers: list of numbers
    :param round_digits_min: minimum number of digits to round
    :return: list of percentages
    """

    def max_precision():
        max_p = 0
        for n in numbers:
            str_n = str(n)
            if '.' in str_n:
                n_precision = len(str_n.split('.')[1])
                max_p = n_precision if n_precision > max_p else max_p
        return max(max_p, round_digits_min)

    if isinstance(numbers, list):
        numbers = np.array(numbers)

    perc = np.round(100 * numbers / np.sum(numbers), max_precision())
    round_max = 99.9
    while np.sum(perc) > 99.99:
        perc = np.round(round_max * numbers / np.sum(numbers), max_precision())
        round_max -= 0.1
    return perc


# From https://stackoverflow.com/questions/3232943/update-value-of-a-nested-dictionary-of-varying-depth
def deep_update(source, overrides) -> list:
    """
    Update a nested dictionary or similar mapping.
    Modify ``source`` in place.
    """
    for key, value in overrides.items():
        if isinstance(value, collections.abc.Mapping) and value:
            returned = deep_update(source.get(key, {}), value)
            source[key] = returned
        else:
            source[key] = overrides[key]
    return source


def convert_data_and_get_series_name(data: pd.DataFrame, field: str) -> tuple[pd.DataFrame, str]:
    """ Convert data to a format that can be used by the API and get the series name of a field.
    :param data: data to convert
    :param field: field to get the series name
    :return: converted data and the converted series name
    """
    converted_data = pd.DataFrame(convert_input_data_to_db_items(data))

    converted_data_columns = converted_data.columns.to_list()
    data_columns = data.columns.to_list()
    series_name = converted_data_columns[data_columns.index(field)]

    return converted_data, series_name


def get_uuids_from_dict(_dict: dict) -> list[str]:
    """ Get all uuids from a dictionary. They follow the pattern '#{'id'}'. """
    uuids = []
    for k, v in _dict.items():
        if isinstance(v, dict):
            uuids.extend(get_uuids_from_dict(v))
        elif isinstance(v, list):
            uuids.extend(get_uuids_from_list(v))
        elif isinstance(v, str) and v.startswith('#{') and v.endswith('}'):
            uuids.append(v[2:-1])
    return uuids


def get_uuids_from_list(_list: list) -> list[str]:
    """ Get all uuids from a list. They follow the pattern '#{'id'}'. """
    uuids = []
    for v in _list:
        if isinstance(v, dict):
            uuids.extend(get_uuids_from_dict(v))
        elif isinstance(v, list):
            uuids.extend(get_uuids_from_list(v))
        elif isinstance(v, str) and v.startswith('#{') and v.endswith('}'):
            uuids.append(v[2:-1])
    return uuids


def get_data_references_from_dict(_dict: dict, previous_keys: Optional[list[Union[str, int]]] = None) -> \
        list[list[str]]:
    """ Get all data references from a dictionary. They follow the pattern '#set_data#'. """
    if previous_keys is None:
        previous_keys = []
    entries = []
    for k, v in _dict.items():
        if isinstance(v, dict):
            entries.extend(get_data_references_from_dict(v, previous_keys=previous_keys + [k]))
        elif isinstance(v, list):
            entries.extend(get_data_references_from_list(v, previous_keys=previous_keys + [k]))
        elif v == '#set_data#':
            entries.append(previous_keys + [k])
    return entries


def get_data_references_from_list(_list: list, previous_keys: Optional[list[Union[int, str]]] = None
                                  ) -> list[list[str]]:
    """ Get all data references from a list. They follow the pattern '#set_data#'. """
    if previous_keys is None:
        previous_keys = []
    entries = []
    for i, v in enumerate(_list):
        if isinstance(v, dict):
            entries.extend(get_data_references_from_dict(v, previous_keys=previous_keys + [i]))
        elif isinstance(v, list):
            entries.extend(get_data_references_from_list(v, previous_keys=previous_keys + [i]))
        elif v == '#set_data#':
            entries.append(previous_keys + [i])
    return entries


def validate_data_is_pandarable(data: Union[str, DataFrame, list[dict], dict]) -> DataFrame:
    """"""
    if isinstance(data, DataFrame):
        df_ = data.copy()
    elif isinstance(data, list):
        try:
            df_ = DataFrame(data)
        except Exception:
            raise ValueError(
                'The data you passed is a list that must be '
                'able to be converted into a pandas dataframe'
            )
    elif isinstance(data, dict):
        try:
            df_ = DataFrame(data)
        except Exception:
            try:
                df_ = DataFrame(data, index=[0])
            except Exception:
                raise ValueError(
                    'The data you passed is a dict that must be '
                    'able to be converted into a pandas dataframe'
                )
    elif isinstance(data, str):
        try:
            d: list[dict] = json.loads(data)
            df_ = DataFrame(d)
        except Exception:
            raise ValueError(
                'The data you passed is a json that must be '
                'able to be converted into a pandas dataframe'
            )
    else:
        raise ValueError(
            'Input data must be a pandas dataframe, '
            'a json or a list of dictionaries'
        )
    return df_


def validate_table_data(
        self, data: Union[str, DataFrame, list[dict], dict], elements: list[str],
):
    """"""
    df_: DataFrame = self._validate_data_is_pandarable(data)

    cols = df_.columns
    try:
        assert all([element in cols for element in elements])
    except AssertionError:
        raise ValueError(
            'Some column names you are specifying '
            'are not in the input dataframe'
        )

    try:
        len_df_: int = len(df_)
        assert all([
            len_df_ == len(df_[~df_[element].isna()])
            for element in elements
        ])
    except AssertionError:
        raise ValueError(
            f'Some of the variables {elements} have none values'
        )


def validate_tree_data(
        self, data: Union[str, list[dict]], vals: list[str],
):
    """To validate Tree and Treemap data"""
    if isinstance(data, list):
        pass
    elif isinstance(data, dict):
        pass
    elif isinstance(data, str):
        data = json.loads(data)
    else:
        raise ValueError('data must be either a list, dict or a json')

    try:
        assert sorted(data.keys()) == sorted(vals)
    except AssertionError:
        raise ValueError('data keys must be "name", "value" and "children"')


def validate_input_form_data(self, data: dict):
    try:
        assert type(data) == dict
    except AssertionError:
        raise ValueError('data must be a dict')

    try:
        assert 'fields' in data
    except AssertionError:
        raise ValueError('"fields" is not a key in the input data')

    try:
        assert type(data['fields']) == list
    except AssertionError:
        raise ValueError('fields must be a list')

    try:
        assert all(['fields' in field_ for field_ in data['fields']])
    except AssertionError:
        raise ValueError('"fields" are not keys in the input data')

    try:
        assert all([
            'fieldName' in field__ and 'mapping' in field__
            for field_ in data['fields']
            for field__ in field_['fields']
        ])
    except AssertionError:
        raise ValueError('"fieldName" and "mapping" are not keys in the input data')


def is_report_data_empty(report_data: Union[list[dict], str, DataFrame, dict, list]) -> bool:
    if isinstance(report_data, DataFrame):
        if report_data.empty:
            return True
        else:
            return False
    elif isinstance(report_data, list) or isinstance(report_data, dict):
        if report_data:
            return False
        else:
            return True
    elif isinstance(report_data, str):
        report_data_: list[dict] = json.loads(report_data)
        if report_data_:
            return False
        else:
            return True
    else:
        raise ValueError(
            f'Data must be a dictionary, JSON or pandas DataFrame '
            f'Provided: {type(report_data)}'
        )


def add_sorting_to_df(df: DataFrame, sort: Optional[dict] = None) -> tuple[DataFrame, dict]:
    """ Add sorting to the data frame. If no sorting is provided the a column named 'sort_values' is added.
    :param df: the data frame
    :param sort: the sorting
    :return: the data frame with sorting and the sorting
    """
    if 'sort_values' not in df.columns:
        df['sort_values'] = range(len(df))
    sort = {'field': 'sort_values', 'direction': 'asc'} if not sort else sort
    df = df[[col for col in df.columns.tolist() if col != 'sort_values'] + ['sort_values']]
    return df, sort


class ShimokuPalette(Enum):
    """Enum for color variables"""
    SUCCESS = 'var(--color-success)'
    SUCCESS_LIGHT = 'var(--color-success-light)'
    WARNING = 'var(--color-warning)'
    WARNING_LIGHT = 'var(--color-warning-light)'
    ERROR = 'var(--color-error)'
    ERROR_LIGHT = 'var(--color-error-light)'
    STATUS_ERROR = 'var(--color-status-error)'
    WHITE = 'var(--color-white)'
    BLACK = 'var(--color-black)'
    GRAY = 'var(--color-gray)'
    BASE_ICON = 'var(--color-base-icon)'
    BACKGROUND = 'var(--background-default)'
    BACKGROUND_PAPER = 'var(--background-paper)'
    PRIMARY = 'var(--color-primary)'
    PRIMARY_LIGHT = 'var(--color-primary-light)'
    PRIMARY_DARK = 'var(--color-primary-dark)'
    SECONDARY = 'var(--color-secondary)'
    SECONDARY_LIGHT = 'var(--color-secondary-light)'
    SECONDARY_DARK = 'var(--color-secondary-dark)'
    STRIPE = 'var(--color-stripe)'
    STRIPE_LIGHT = 'var(--color-stripe-light)'
    COM_RED = 'var(--complementary-red)'
    COM_RED_LIGHT = 'var(--complementary-red-light)'
    COM_GREEN = 'var(--complementary-green)'
    COM_YELLOW = 'var(--complementary-yellow)'
    COM_ORANGE = 'var(--complementary-orange)'
    COM_AQUA = 'var(--complementary-aqua)'
    COM_VIOLET = 'var(--complementary-violet)'
    CHART_C1 = 'var(--chart-C1)'
    CHART_C2 = 'var(--chart-C2)'
    CHART_C3 = 'var(--chart-C3)'
    CHART_C4 = 'var(--chart-C4)'
    CHART_C5 = 'var(--chart-C5)'
    CHART_C6 = 'var(--chart-C6)'
    CHART_C7 = 'var(--chart-C7)'
    CHART_C8 = 'var(--chart-C8)'
    CHART_C9 = 'var(--chart-C9)'
    CHART_C10 = 'var(--chart-C10)'


def interpret_color(color_def: Union[list, str, int]) -> Union[str, dict]:
    def rbg_to_hex(r: int, g: int, b: int) -> str:
        def clamp(x: int) -> int:
            return max(0, min(x, 255))

        # from https://stackoverflow.com/questions/3380726/converting-an-rgb-color-tuple-to-a-hexidecimal-string
        return "#{0:02x}{1:02x}{2:02x}".format(clamp(r), clamp(g), clamp(b))

    # from https://xkcd.com/color/rgb/
    color_defs = {
        "purple": "#7e1e9c",
        "red": "#e50000",
        "green": "#15b01a",
        "blue": "#0343df",
        "pink": "#ff81c0",
        "brown": "#653700",
        "orange": "#f97306",
        "yellow": "#ffff14",
        "gray": "#929591",
        "violet": "#9a0eea",
        "cyan": "#00ffff",
        "success": ShimokuPalette.SUCCESS.value,
        "success-light": ShimokuPalette.SUCCESS_LIGHT.value,
        "warning": ShimokuPalette.WARNING.value,
        "warning-light": ShimokuPalette.WARNING_LIGHT.value,
        "error": ShimokuPalette.ERROR.value,
        "error-light": ShimokuPalette.ERROR_LIGHT.value,
        "status-error": ShimokuPalette.STATUS_ERROR.value,
        "white": ShimokuPalette.WHITE.value,
        "black": ShimokuPalette.BLACK.value,
        "base-icon": ShimokuPalette.BASE_ICON.value,
        "background": ShimokuPalette.BACKGROUND.value,
        "background-paper": ShimokuPalette.BACKGROUND_PAPER.value,
        "primary": ShimokuPalette.PRIMARY.value,
        "primary-light": ShimokuPalette.PRIMARY_LIGHT.value,
        "primary-dark": ShimokuPalette.PRIMARY_DARK.value,
        "main": ShimokuPalette.CHART_C1.value,
        "secondary": ShimokuPalette.SECONDARY.value,
        "secondary-light": ShimokuPalette.SECONDARY_LIGHT.value,
        "secondary-dark": ShimokuPalette.SECONDARY_DARK.value,
        "active": ShimokuPalette.CHART_C2.value,
        "caution": ShimokuPalette.CHART_C3.value,

    }

    if isinstance(color_def, int):
        return f'var(--chart-C{abs(color_def)})'
    elif isinstance(color_def, (list, tuple)):
        color_def = rbg_to_hex(color_def[0], color_def[1], color_def[2])
    elif color_def in color_defs:
        color_def = color_defs[color_def]

    return color_def


def transform_dict_js_to_py(options_str: str):
    """https://discuss.dizzycoding.com/how-to-convert-raw-javascript-object-to-python-dictionary/"""
    options_str = options_str.replace('\n', '')
    options_str = options_str.replace(';', '')
    return json5.loads(options_str)


def retrieve_data_from_options(options: dict) -> Union[dict, list]:
    """Retrieve data from eCharts options

    Example
    -----------
    input options = {'title': {'text': 'Stacked Area Chart'},
         'tooltip': {'trigger': 'axis',
          'axisPointer': {'type': 'cross', 'label': {'backgroundColor': '#6a7985'}}},
         'legend': {'data': ['Email',
           'Union Ads',
           'Video Ads',
           'Direct',
           'Search Engine']},
         'toolbox': {'feature': {'saveAsImage': {}}},
         'grid': {'left': '3%', 'right': '4%', 'bottom': '3%', 'containLabel': True},
         'xAxis': [{'type': 'category',
           'boundaryGap': False,
           'data': ['Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun']}],
         'yAxis': [{'type': 'value'}],
         'series': [{'name': 'Email',
           'type': 'line',
           'stack': 'Total',
           'areaStyle': {},
           'emphasis': {'focus': 'series'},
           'data': [120, 132, 101, 134, 90, 230, 210]},
          {'name': 'Union Ads',
           'type': 'line',
           'stack': 'Total',
           'areaStyle': {},
           'emphasis': {'focus': 'series'},
           'data': [220, 182, 191, 234, 290, 330, 310]},
          {'name': 'Video Ads',
           'type': 'line',
           'stack': 'Total',
           'areaStyle': {},
           'emphasis': {'focus': 'series'},
           'data': [150, 232, 201, 154, 190, 330, 410]},
          {'name': 'Direct',
           'type': 'line',
           'stack': 'Total',
           'areaStyle': {},
           'emphasis': {'focus': 'series'},
           'data': [320, 332, 301, 334, 390, 330, 320]},
          {'name': 'Search Engine',
           'type': 'line',
           'stack': 'Total',
           'label': {'show': True, 'position': 'top'},
           'areaStyle': {},
           'emphasis': {'focus': 'series'},
           'data': [820, 932, 901, 934, 1290, 1330, 1320]}]
        }

    output
        [{'Mon': 120,
          'Tue': 132,
          'Wed': 101,
          'Thu': 134,
          'Fri': 90,
          'Sat': 230,
          'Sun': 210},
         {'Mon': 220,
          'Tue': 182,
          'Wed': 191,
          'Thu': 234,
          'Fri': 290,
          'Sat': 330,
          'Sun': 310},
         {'Mon': 150,
          'Tue': 232,
          'Wed': 201,
          'Thu': 154,
          'Fri': 190,
          'Sat': 330,
          'Sun': 410},
         {'Mon': 320,
          'Tue': 332,
          'Wed': 301,
          'Thu': 334,
          'Fri': 390,
          'Sat': 330,
          'Sun': 320},
         {'Mon': 820,
          'Tue': 932,
          'Wed': 901,
          'Thu': 934,
          'Fri': 1290,
          'Sat': 1330,
          'Sun': 1320
        }
    ]
    """
    rows = []
    data = []
    cols = []
    if 'xAxis' in options:
        if type(options['xAxis']) == list:
            if len(options['xAxis']) == 1:
                if 'data' in options['xAxis'][0]:
                    rows = options['xAxis'][0]['data']
        elif type(options['xAxis']) == dict:
            if 'data' in options['xAxis']:
                rows = options['xAxis']['data']
            elif type(options['xAxis']) == dict:
                if 'data' in options['yAxis']:
                    rows = options['yAxis']['data']
        else:
            raise ValueError('xAxis has multiple values only 1 allowed')
    elif 'radar' in options:
        if 'indicator' in options['radar']:
            rows = [element['name'] for element in options['radar']['indicator']]
        elif type(options['radar']) == dict:
            raise NotImplementedError('Multi-radar not implemented')

    if 'data' in options:
        data = options['data']
    if 'series' in options:
        if 'data' in options['series']:
            pass
        else:
            for serie in options['series']:
                if 'data' in serie:
                    if serie.get('type') in ['pie', 'gauge']:
                        for datum in serie['data']:
                            data.append(datum)
                    elif serie.get('type') == 'radar':
                        for datum in serie['data']:
                            data.append(datum['value'])
                            cols.append(datum['name'])
                        break
                    else:
                        data.append(serie['data'])

                if 'name' in serie:
                    cols.append(serie['name'])
                elif 'type' in serie:
                    cols.append(serie['type'])
    else:
        return {}

    df = DataFrame(data)
    if not rows and not cols:
        return df.reset_index().to_dict(orient='records')
    if not rows:
        return df.to_dict(orient='records')

    if rows:
        df.columns = rows
    df_ = df.T
    df_.columns = cols
    return df_.reset_index().to_dict(orient='records')




import tqdm
from typing import Optional, List

from shimoku.api.resources.business import Business

from shimoku.utils import create_function_name, create_normalized_name

from shimoku.code_gen.file_generator import CodeGenFileHandler
from shimoku.code_gen.business_code_gen.apps_code_gen.code_gen_from_apps import AppCodeGen
from shimoku.code_gen.utils_code_gen import code_gen_from_list, code_gen_from_dict

from shimoku.execution_logger import ClassWithLogging
import logging
logger = logging.getLogger(__name__)


class BusinessCodeGen(ClassWithLogging):

    _module_logger = logger

    def __init__(self, business: Business, output_path: str):
        self._business = business
        self._output_path = f'{output_path}/{create_function_name(business["name"])}'
        self._file_generator = CodeGenFileHandler(self._output_path)

    async def generate_code(self, business_id: str, menu_paths: List[str], pbar: Optional[tqdm.tqdm] = None):
        """ Use the resources in the API to generate code_lines for the SDK. Create a file in
        the specified path with the generated code_lines.
        :param business_id: business id to use
        :param menu_paths: list of menu paths to generate code for
        :param pbar: progress bar to use
        """
        import_code_lines: List[str] = [
            'from shimoku import Client',
        ]
        main_code_lines: List[str] = [f'shimoku_client.set_workspace("{business_id}")', '']

        if self._business['theme']:
            theme_code_lines = code_gen_from_dict(self._business['theme'], deep=4)
            main_code_lines.extend([
                f'shimoku_client.workspaces.update_workspace("{business_id}", theme={theme_code_lines[0][4:]})',
                *theme_code_lines[1:],
                ''
            ])

        if menu_paths:
            menu_paths = [create_normalized_name(menu_path) for menu_path in menu_paths]
        apps = [app for app in sorted(await self._business.get_apps(), key=lambda x: x['order'])
                if menu_paths is None or app['normalizedName'] in menu_paths]

        extra_apps_for_dashboard = {}
        last_dashboard = None
        dashboards = sorted(await self._business.get_dashboards(), key=lambda x: x['order'])
        created_dashboards = []
        for app in apps:
            app_id = app['id']
            app_code_gen = AppCodeGen(app, self._output_path, pbar=pbar)

            await app_code_gen.generate_code()

            import_code_lines.append(f'from .{app_code_gen.app_f_name}.app import {app_code_gen.app_f_name}')
            aux_code_lines = []
            try:
                first_dashboard, *extra_dashboards = [dashboard
                                                      for dashboard in dashboards
                                                      if app_id in await dashboard.list_app_ids()]
            except IndexError:
                self._module_logger.warning(f'App {app["name"]} has no dashboards')
                continue
            created_dashboards.append(first_dashboard)
            if first_dashboard['name'] != last_dashboard:
                aux_code_lines.append(f'shimoku_client.set_board("{first_dashboard["name"]}")')
                last_dashboard = first_dashboard['name']
            for dashboard in extra_dashboards:
                extra_apps_for_dashboard.setdefault(dashboard['name'], []).append(app['name'])

            aux_code_lines.append(f'{app_code_gen.app_f_name}(shimoku_client)')
            main_code_lines.extend(aux_code_lines)

        for dashboard_name, app_names in extra_apps_for_dashboard.items():
            app_names_code_lines = code_gen_from_list(app_names, deep=4)
            main_code_lines.extend([
                '',
                'shimoku_client.boards.group_menu_paths('
                '    name="' + dashboard_name + '",',
                f'    menu_path_names={app_names_code_lines[0][4:]}',
                *app_names_code_lines[1:],
                ')'
            ])

        if created_dashboards != dashboards:
            dashboards_code_lines = code_gen_from_list([dashboard['name'] for dashboard in dashboards], deep=4)
            main_code_lines.extend([
                '',
                'shimoku_client.workspaces.change_boards_order(',
                f'   uuid="{business_id}",',
                f'   boards={dashboards_code_lines[0][4:]}',
                *dashboards_code_lines[1:],
                ')'
            ])

        business_name = create_function_name(self._business['name'])
        self._file_generator.generate_script_file(
            f'workspace_{business_name}',
            [
                *import_code_lines,
                '',
                '',
                f'def workspace_{business_name}(shimoku_client: Client):',
                *['    ' + line for line in main_code_lines],
                '',
            ]
        )
        # Create an __init__.py file for the imports to work
        self._file_generator.generate_script_file('__init__', [''])



from typing import Optional, TYPE_CHECKING

import asyncio

from shimoku.api.base_resource import Resource, ResourceCache
from shimoku.api.resources.role import Role, create_role, get_role, get_roles, delete_role
from shimoku.api.resources.app import App

if TYPE_CHECKING:
    from shimoku.api.resources.business import Business
from shimoku.exceptions import BoardError
import logging
from shimoku.execution_logger import log_error
logger = logging.getLogger(__name__)


class Dashboard(Resource):
    """ Dashboard resource class """

    resource_type = 'dashboard'
    alias_field = 'name'
    plural = 'dashboards'

    class AppDashboard(Resource):
        """ AppDashboard resource class """

        resource_type = 'appDashboard'
        plural = 'appDashboards'

        def __init__(self, parent: 'Dashboard', uuid: Optional[str] = None, app: App = None,
                     db_resource: Optional[dict] = None):

            params = dict(
                appId=app['id'] if app else None,
            )
            super().__init__(parent=parent, uuid=uuid, db_resource=db_resource,
                             check_params_before_creation=['appId'], params=params)

        async def delete(self):
            """ Deletes the appDashboard """
            return await self._base_resource.delete()

    def __init__(self, parent: 'Business', uuid: Optional[str] = None, alias: Optional[str] = None,
                 db_resource: Optional[dict] = None):

        params = dict(
            name=alias,
            order=0,
            isDisabled=False,
            publicPermission=dict(
                isPublic=False,
                permission='READ',
                token='default token'
            )
        )

        super().__init__(parent=parent, uuid=uuid, db_resource=db_resource,
                         children=[Dashboard.AppDashboard, Role],
                         check_params_before_creation=['name'], params=params)

        self.currently_in_use = False

    async def delete(self):
        """ Deletes the dashboard """
        if self.currently_in_use:
            log_error(logger, f'Workspace {str(self)} is currently in use and cannot be deleted', BoardError)
        return await self._base_resource.delete()

    async def update(self):
        """ Updates the dashboard """
        return await self._base_resource.update()

    async def list_app_ids(self) -> list[str]:
        """ Returns the list of app ids in the dashboard """
        cache: ResourceCache = self._base_resource.children[Dashboard.AppDashboard]
        cache_list = await cache.list()
        return list(set([app_dashboard['appId'] for app_dashboard in cache_list]))

    async def insert_app(self, app: App):
        """ Inserts an app in the dashboard
        :param app: The App to insert
        """
        dashboards = await self._base_resource.parent.get_dashboards()
        app_ids_lists = await asyncio.gather(*[dashboard.list_app_ids() for dashboard in dashboards])
        app_ids = [app_id for app_ids_list in app_ids_lists for app_id in app_ids_list]
        if app['id'] in app_ids:
            logger.warning(f"Menu path {str(app)} already exists in another board, it will appear in both boards")
        await self._base_resource.create_child(Dashboard.AppDashboard, appId=app['id'])
        logger.info(f"Menu path {str(app)} added to board {str(self)}")

    async def remove_app(self, app: App):
        """ Removes an app from the dashboard
        :param app: The App to remove
        """
        tasks = []
        cache: ResourceCache = self._base_resource.children[Dashboard.AppDashboard]
        cache_list = await cache.list()
        for app_dashboard in cache_list:
            if app_dashboard['appId'] == app['id']:
                tasks.append(cache.delete(uuid=app_dashboard['id']))

        await asyncio.gather(*tasks)
        logger.info(f"Menu path {str(app)} removed from board {str(self)}")

    async def remove_all_apps(self):
        """ Removes all apps from the dashboard """
        tasks = []
        cache: ResourceCache = self._base_resource.children[Dashboard.AppDashboard]
        cache_list = await cache.list()
        for app_dashboard in cache_list:
            tasks.append(cache.delete(uuid=app_dashboard['id']))

        await asyncio.gather(*tasks)

    # Role methods
    get_role = get_role
    get_roles = get_roles
    create_role = create_role
    delete_role = delete_role



from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from shimoku.code_gen.business_code_gen.apps_code_gen.code_gen_from_apps import AppCodeGen
    from shimoku.api.resources.report import Report


async def code_gen_from_filter(
        self: 'AppCodeGen', report: 'Report'
) -> list[str]:
    """ Generate code for a filter report.
    :param report: report to generate code from
    :return: list of code lines
    """
    filter_def = report['properties']['filter'][0]
    mapping = report['properties']['mapping'][0]
    field_name = filter_def['field']
    data_set = await self._app.get_data_set(mapping['id'])
    field = field_name if data_set['columns'] else mapping[field_name]
    report_params = [
        f'    order={report["order"]},',
        f'    data="{data_set["name"]}",',
        f'    field="{field}",',
    ]
    if report['sizeColumns'] != 4:
        report_params.append(f'    cols_size={report["sizeColumns"]},')
    if report['sizeRows'] != 1:
        report_params.append(f'    rows_size={report["sizeRows"]},')
    if report['sizePadding'] != '0,0,0,0':
        report_params.append(f'    padding="{report["sizePadding"]}",')
    if filter_def['inputType'] == 'CATEGORICAL_MULTI':
        report_params.append(f'    multi_select=True,')
    return [
        'shimoku_client.plt.filter(',
        *report_params,
        ')'
    ]



from typing import Optional

import logging
from shimoku.execution_logger import ClassWithLogging
from shimoku.api.resources.app import App
from shimoku.api.resources.report import Report
logger = logging.getLogger(__name__)


class ComponentsLayer(ClassWithLogging):
    """
    This class is used to interact with the API at the component level.
    """

    _module_logger = logger
    _use_info_logging = True
    
    def __init__(self, app: Optional[App]):
        self._app = app
    
    async def get_components_in_sub_path(self, path: str) -> list[dict]:
        """
        :param path: path to the components
        :return: list of components
        """
        return [report.cascade_to_dict() for report in (await self._app.get_reports()) if report['path'] == path]

    async def get_component(self, uuid: str) -> dict:
        """
        :param uuid: uuid of the component
        :return: component
        """
        return (await self._app.get_report(uuid=uuid)).cascade_to_dict()

    async def delete_component(self, uuid: str):
        """
        :param uuid: uuid of the component
        """
        await self._app.delete_report(uuid=uuid)

    async def get_component_data_set_links(self, uuid: str) -> list[dict]:
        """
        :param uuid: uuid of the component
        :return: list of data set links
        """
        report: Report = await self._app.get_report(uuid=uuid)
        return [rds.cascade_to_dict() for rds in (await report.get_report_data_sets())]



from ...report import Report


class Button(Report):
    report_type = 'BUTTON'

    default_properties = dict(
        **Report.default_properties,
        text='More Info',
        align='stretch',
        events={}
    )



from typing import Optional, Union, TYPE_CHECKING

from pandas import DataFrame

from shimoku.plt.EChart_definitions.default_echart_options import (
    get_common_echart_options, get_common_series_options, InvertibleDictFunction, generic_echart_function_params
)
if TYPE_CHECKING:
    from shimoku.plt.plt_layer import PlotLayer


from shimoku.utils import get_arg_names
common_opt = get_common_echart_options()


def get_pie_chart_options_function() -> InvertibleDictFunction:
    pie_func = InvertibleDictFunction(pie_chart)
    pie_func.set_items(get_common_echart_options())
    pie_func.delete_items(['yAxis', 'xAxis', 'grid'])
    pie_func.delete_items(offset=['toolbox', 'feature'], keys=['magicType'])
    pie_func.set_items(
        final=True, offset=['series', 0],
        items={
            'type': 'pie',
            'data': '#set_data#',
            'itemStyle': {
                'borderWidth': 0,
                'borderRadius': 5
            },
            'radius': '70%'
        },
    )
    pie_func.custom_operation(
        func=lambda params:
            params.update(
                {'fields': [{'name': pie_func.params['names'], 'value': pie_func.params['values']}]}
            ),
        inverse_func=lambda params:
            (params.update({
                'names': pie_func.params['fields'][0]['name'],
                'values': pie_func.params['fields'][0]['value']}
            ), params.pop('fields'))
    )
    pie_func.params_to_params(generic_echart_function_params, replace=True)
    return pie_func


async def pie_chart(
    self: 'PlotLayer', order: int, names: str, values: str,
    data: Union[str, DataFrame, list[dict]],
    title: Optional[str] = None, rows_size: Optional[int] = 2,
    cols_size: Optional[int] = 3, padding: Optional[str] = None,
    option_modifications: Optional[dict] = None,
):
    """ Create a pie chart """
    local_vars = locals()
    kwargs = {k: local_vars[k] for k in get_arg_names(pie_chart) if k != 'self' and k in local_vars}
    pie_func = get_pie_chart_options_function()
    await self._create_echart(**pie_func(**kwargs))


async def doughnut_chart(
    self: 'PlotLayer', order: int, names: str, values: str,
    data: Union[str, DataFrame, list[dict]],
    title: Optional[str] = None, rows_size: Optional[int] = 2,
    cols_size: Optional[int] = 3, padding: Optional[str] = None,
    option_modifications: Optional[dict] = None,
):
    """ Create a doughnut chart """
    common_options = get_common_echart_options()
    series_options = get_common_series_options()
    del common_options['yAxis']
    del common_options['xAxis']
    del common_options['grid']
    del common_options['toolbox']['feature']['magicType']
    common_options['tooltip']['trigger'] = 'item'
    series_options.update({
        'type': 'pie',
        'data': '#set_data#',
        'itemStyle': {'borderWidth': 0, 'borderRadius': 5},
        'radius': ['40%', '70%'],
        'label': {'show': False, 'position': 'center'},
        'emphasis': {'label': {'show': True, 'fontSize': '30', 'fontWeight': 'bold'}},
        'labelLine': {'show': False}
    })
    common_options['series'] = [series_options]

    await self._create_echart(
        fields=[{'name': names, 'value': values}], options=common_options,
        data_mapping_to_tuples=await self._choose_data(order, data), option_modifications=option_modifications,
        order=order, title=title, rows_size=rows_size, cols_size=cols_size, padding=padding)


async def rose_chart(
    self: 'PlotLayer', order: int, names: str, values: str,
    data: Union[str, DataFrame, list[dict]],
    title: Optional[str] = None, rows_size: Optional[int] = 2,
    cols_size: Optional[int] = 3, padding: Optional[str] = None,
    option_modifications: Optional[dict] = None,
):
    """ Create a rose chart """
    common_options = get_common_echart_options()
    series_options = get_common_series_options()
    del common_options['yAxis']
    del common_options['xAxis']
    del common_options['grid']
    del common_options['toolbox']['feature']['magicType']
    common_options['tooltip']['trigger'] = 'item'
    series_options.update({
        'type': 'pie',
        'roseType': 'area',
        'data': '#set_data#',
        'itemStyle': {'borderWidth': 0, 'borderRadius': 5},
        'radius': ['10%', '70%'],
    })
    common_options['series'] = [series_options]

    await self._create_echart(
        fields=[{'name': names, 'value': values}], options=common_options,
        data_mapping_to_tuples=await self._choose_data(order, data), option_modifications=option_modifications,
        order=order, title=title, rows_size=rows_size, cols_size=cols_size, padding=padding,
    )



from shimoku.cli.cloud.list import add_list_parser
from shimoku.cli.cloud.create import add_create_parser
from shimoku.cli.cloud.delete import add_delete_parser
from shimoku.cli.cloud.get import add_get_parser
from shimoku.cli.cloud.update import add_update_parser
from shimoku.cli.cloud.execute import add_execute_parser
from shimoku.cli import CLIParser
from typing import Optional


def add_cloud_parser(parser: Optional[CLIParser] = None):
    """ Function to add the cloud parser to a parser
    :param parser: Parser to add the cloud parser to
    :return: Cloud parser
    """
    params = {
        'name': 'cloud',
        'description': 'Commands to interact with the cloud',
    }

    if parser:
        cloud_parser = parser.add_command(**params)
    else:
        cloud_parser = CLIParser(**params)

    add_list_parser(cloud_parser)
    add_get_parser(cloud_parser)
    add_create_parser(cloud_parser)
    add_delete_parser(cloud_parser)
    add_update_parser(cloud_parser)
    add_execute_parser(cloud_parser)

    return cloud_parser


if __name__ == "__main__":
    add_cloud_parser().parse_args()



from typing import TYPE_CHECKING, List
from shimoku.utils import create_function_name
from shimoku.code_gen.business_code_gen.apps_code_gen.reports_code_gen. \
        report_types_code_gen.code_gen_from_tabs import code_gen_tabs_and_other

if TYPE_CHECKING:
    from shimoku.code_gen.business_code_gen.apps_code_gen.code_gen_from_apps import AppCodeGen


async def code_gen_from_reports_tree(self: 'AppCodeGen', path: str) -> List[str]:
    code_lines = [
        # *(await code_gen_tabs_functions(all_tab_groups[path]) if path in all_tab_groups else []),
        # *(await code_gen_modals_functions(tree[path]['modals']) if path in tree else []),
    ]
    tree = self._code_gen_tree.tree
    for modal_dict in tree[path]['modals']:
        code_lines.extend([
            '',
            f'modal_{create_function_name(modal_dict["name"])}(shimoku_client)'
        ])
    # if len(tree[path]['modals']) > 0:
    #     code_lines.extend(['', 'shimoku_client.plt.pop_out_of_modal()', ''])
    code_lines.extend(['', *await code_gen_tabs_and_other(self, tree[path])])
    return code_lines



from __future__ import absolute_import

from typing import Dict, Optional

from shimoku.async_execution_pool import AutoAsyncExecutionPool, decorate_class_to_auto_async

from shimoku.api.client import ApiClient
from shimoku.api.user_access_classes.universes_layer import UniversesLayer, Universe
from shimoku.api.user_access_classes.generated_headers.UniversesLayerHeader import UniversesLayerHeader
from shimoku.api.user_access_classes.businesses_layer import WorkspacesLayer, Business
from shimoku.api.user_access_classes.generated_headers.WorkspacesLayerHeader import WorkspacesLayerHeader
from shimoku.api.user_access_classes.activity_templates_layer import ActivityTemplatesLayer
from shimoku.api.user_access_classes.generated_headers.ActivityTemplatesLayerHeader import ActivityTemplatesLayerHeader
from shimoku.api.user_access_classes.dashboards_layer import BoardsLayer, Dashboard
from shimoku.api.user_access_classes.generated_headers.BoardsLayerHeader import BoardsLayerHeader
from shimoku.api.user_access_classes.apps_layer import MenuPathsLayer, App
from shimoku.api.user_access_classes.generated_headers.MenuPathsLayerHeader import MenuPathsLayerHeader
from shimoku.api.user_access_classes.reports_layer import ComponentsLayer
from shimoku.api.user_access_classes.generated_headers.ComponentsLayerHeader import ComponentsLayerHeader
from shimoku.api.user_access_classes.data_sets_layer import DataSetsLayer
from shimoku.api.user_access_classes.generated_headers.DataSetsLayerHeader import DataSetsLayerHeader
from shimoku.api.user_access_classes.files_layer import FilesLayer
from shimoku.api.user_access_classes.generated_headers.FilesLayerHeader import FilesLayerHeader
from shimoku.api.user_access_classes.activities_layer import ActivitiesLayer
from shimoku.api.user_access_classes.generated_headers.ActivitiesLayerHeader import ActivitiesLayerHeader

from shimoku.plt.plt_layer import PlotLayer
from shimoku.plt.generated_headers.PlotLayerHeader import PlotLayerHeader
from shimoku.plt.utils import create_normalized_name, ShimokuPalette

from shimoku.ai.ai_layer import AILayer
from shimoku.ai.generated_headers.AILayerHeader import AILayerHeader

from shimoku.utils import IN_BROWSER_PYODIDE, EventType

from shimoku.exceptions import WorkspaceError, BoardError

import shimoku_components_catalog.html_components

import logging
from shimoku.execution_logger import log_error, configure_logging, logging_before_and_after
logger = logging.getLogger(__name__)


class Client:

    def __init__(
        self, universe_id: str = 'local', environment: str = 'production',
        access_token: Optional[str] = None, config: Optional[Dict] = None,
        verbosity: str = None, async_execution: bool = False,
        local_port: int = 8000, retry_attempts: int = 5
    ):
        self.playground: bool = universe_id == 'local' and not access_token
        if self.playground:
            access_token = 'local'
        if universe_id == 'local' and not self.playground:
            log_error(logger, 'Local universe can only be used in playground mode.', AttributeError)
        self.access_token = access_token
        self.environment = environment
        self._access_token = access_token
        self.universe_id = universe_id
        self.workspace_id = None
        self.board_id = None
        self.menu_path_id = None
        if not config:
            config = {}

        self.server_host = '127.0.0.1'
        self.local_port = local_port

        self.configure_logging = configure_logging
        if verbosity:
            self.configure_logging(verbosity)

        if access_token and access_token != "":
            config = {'access_token': access_token}

        self._api_client = ApiClient(
            config=config, environment=environment, playground=self.playground,
            server_host=self.server_host, server_port=local_port, retry_attempts=retry_attempts)

        self._universe_object = Universe(self._api_client, uuid=universe_id)
        self._business_object: Optional[Business] = None
        self._app_object: Optional[App] = None
        self._dashboard_object: Optional[Dashboard] = None

        self._async_pool = AutoAsyncExecutionPool(api_client=self._api_client)

        if async_execution:
            self.activate_async_execution()
        else:
            self.activate_sequential_execution()

        self.universes: UniversesLayerHeader = (
            decorate_class_to_auto_async(UniversesLayer, self._async_pool)(self._api_client))
        self.workspaces: WorkspacesLayerHeader = (
            decorate_class_to_auto_async(WorkspacesLayer, self._async_pool)(self._universe_object))
        self.activity_templates: ActivityTemplatesLayerHeader = (
            decorate_class_to_auto_async(ActivityTemplatesLayer, self._async_pool)(self._universe_object))
        self.boards: BoardsLayerHeader = (
            decorate_class_to_auto_async(BoardsLayer, self._async_pool)(self._business_object))
        self.menu_paths: MenuPathsLayerHeader = (
            decorate_class_to_auto_async(MenuPathsLayer, self._async_pool)(self._business_object))
        self.components: ComponentsLayerHeader = (
            decorate_class_to_auto_async(ComponentsLayer, self._async_pool)(self._app_object))
        self.data: DataSetsLayerHeader = (
            decorate_class_to_auto_async(DataSetsLayer, self._async_pool)(self._app_object))
        self.io: FilesLayerHeader = (
            decorate_class_to_auto_async(FilesLayer, self._async_pool)(self._app_object))
        self.activities: ActivitiesLayerHeader = (
            decorate_class_to_auto_async(ActivitiesLayer, self._async_pool)(self._app_object))
        self.ai: AILayerHeader = (
            decorate_class_to_auto_async(AILayer, self._async_pool)(
                self.access_token, self._universe_object, self._app_object))

        self._reuse_data_sets = False
        self._shared_dfs = {}
        self._shared_custom_data = {}
        self.plt: PlotLayerHeader = (
            decorate_class_to_auto_async(PlotLayer, self._async_pool)(
                self._async_pool, self._app_object, self._reuse_data_sets))

        self._async_pool.current_app = self._app_object
        self._async_pool.universe = self._universe_object

        self.html_components = shimoku_components_catalog.html_components

    def set_workspace(
        self, uuid: Optional[str] = None, name: Optional[str] = None
    ) -> None:
        """Set workspace id for the client.
        :param uuid: Workspace uuid
        :param name: Workspace name
        """
        if self._api_client.playground:
            uuid, name = 'local', None

        @logging_before_and_after(logging_level=logger.info, name='set_workspace')
        async def a_set_workspace(_self):
            if self._business_object:
                self._business_object.currently_in_use = False
            business: Optional[Business] = await self._universe_object.get_business(uuid=uuid, name=name)
            if not business:
                log_error(logger, f'Workspace {name if name else uuid} not found.', WorkspaceError)
            self._business_object = business
            self._business_object.currently_in_use = True
            self.workspace_id = self._business_object['id']

            if self._app_object:
                self.plt.raise_if_cant_change_path()
                self._app_object.currently_in_use = False
            self._app_object = None
            self.menu_path_id = None
            if self._dashboard_object:
                self._dashboard_object.currently_in_use = False
            self._dashboard_object = None
            self.board_id = None

            self.boards.__init__(self._business_object)
            self.menu_paths.__init__(self._business_object)
            self.components.__init__(self._app_object)
            self.activities.__init__(self._app_object)
            self.plt.__init__(self._async_pool, self._app_object, self._reuse_data_sets)
            self.data.__init__(self._app_object)
            self.io.__init__(self._app_object)
            self.ai.__init__(self.access_token, self._universe_object, self._app_object)

        return self._async_pool.auto_async_func_call(
            name='set_workspace',
            func_self=self,
            func=a_set_workspace,
        )

    def pop_out_of_menu_path(self) -> None:
        """ Pop out of the menu path. """
        @logging_before_and_after(logging_level=logger.info, name='pop_out_of_menu_path')
        async def a_pop_out_of_menu_path(_self):
            data_names = self.plt.get_shared_data_names()
            if data_names:
                logger.info(f'Shared data entries will no longer be available: {data_names}, set them again if needed.')
            self.plt.clear_context()
            self._app_object.currently_in_use = False
            self._app_object = None
            self.menu_path_id = None

        return self._async_pool.auto_async_func_call(
            name='pop_out_of_menu_path',
            func_self=self,
            func=a_pop_out_of_menu_path,
        )

    def set_board(self, name: str) -> None:
        """Set the board in use for the following apps being called.
        :param name: board name
        """
        if not self._business_object:
            log_error(logger, 'Workspace not set. Please use set_workspace() method first.', AttributeError)
        if self._app_object:
            self.pop_out_of_menu_path()
        if self._dashboard_object:
            self._dashboard_object.currently_in_use = False

        @logging_before_and_after(logging_level=logger.info, name='set_board')
        async def a_set_board(_self):
            self._dashboard_object = await self._business_object.get_dashboard(name=name)
            self._dashboard_object.currently_in_use = True
            self.board_id = self._dashboard_object['id']

        return self._async_pool.auto_async_func_call(
            name='set_board',
            func_self=self,
            func=a_set_board,
        )

    def pop_out_of_board(self) -> None:
        """ Pop out of the dashboard. """
        if not self._business_object:
            log_error(logger, 'Workspace not set. Please use set_workspace() method first.', AttributeError)
        if not self._dashboard_object:
            log_error(logger, 'Board not set. Please use set_board() method first.', BoardError)
        if self._app_object:
            self.pop_out_of_menu_path()

        @logging_before_and_after(logging_level=logger.info, name='pop_out_of_board')
        async def a_pop_out_of_board(_self):
            self._dashboard_object.currently_in_use = False
            self._dashboard_object = None
            self.board_id = None

        return self._async_pool.auto_async_func_call(
            name='pop_out_of_board',
            func_self=self,
            func=a_pop_out_of_board,
        )

    def enable_caching(self):
        """ Enable caching. """
        self._api_client.cache_enabled = True

    def disable_caching(self):
        """ Disable caching. """
        self._api_client.cache_enabled = False

    def reuse_data_sets(self):
        """ Reuse data sets from the api. """
        self._reuse_data_sets = True
        if self._app_object:
            self.plt.reuse_data_sets = True

    def update_data_sets(self):
        """ Update data sets in the api. """
        self._reuse_data_sets = False
        if self._app_object:
            self.plt.reuse_data_sets = False

    async def _change_app(self, menu_path: str, dont_add_to_dashboard: bool):
        """Change app in use for the following calls.
        :param menu_path: Menu path of the app
        :param dont_add_to_dashboard: Whether to add the menu path to the dashboard
        """
        if self._app_object:
            self._app_object.currently_in_use = False

        app: App = await self._business_object.get_app(name=menu_path)
        self._app_object = app
        app.currently_in_use = True
        self.activities.__init__(self._app_object)
        self.components.__init__(self._app_object)
        self.plt.__init__(self._async_pool, self._app_object, self._reuse_data_sets)
        self.data.__init__(self._app_object)
        self.io.__init__(self._app_object)
        self.ai.__init__(self.access_token, self._universe_object, self._app_object)

        self.menu_path_id = self._app_object['id']

        if dont_add_to_dashboard:
            return

        if not self._dashboard_object:
            self._dashboard_object = await self._business_object.get_dashboard(name='Default Name')

        if self._app_object['id'] not in await self._dashboard_object.list_app_ids():
            await self._dashboard_object.insert_app(self._app_object)

        self.menu_path_id = self._app_object['id']

    def set_menu_path(
        self, name: str,
        sub_path: Optional[str] = None,
        dont_add_to_dashboard: bool = False
    ) -> None:
        """Set menu path for the client.
        :param name: Menu path
        :param sub_path: Sub path
        :param dont_add_to_dashboard: Whether to add the menu path to the dashboard
        """
        if not self._business_object:
            log_error(logger, 'Workspace not set. Please use set_workspace() method first.', AttributeError)
        normalized_name = create_normalized_name(name)
        path = sub_path if sub_path else None
        data_names = []
        if self._app_object:
            self.plt.raise_if_cant_change_path()
            if self._app_object['normalizedName'] == normalized_name:
                self.plt.change_path(path)
                return
            data_names = self.plt.get_shared_data_names()

        @logging_before_and_after(logging_level=logger.info, name='set_menu_path')
        async def a_set_menu_path(_self):
            await self._change_app(name, dont_add_to_dashboard)
            self.plt.change_path(path)
            if data_names:
                logger.info(f'Shared data entries will no longer be available: {data_names}, set them again if needed.')

        return self._async_pool.auto_async_func_call(
            name='set_menu_path',
            func_self=self,
            func=a_set_menu_path,
        )

    async def _create_business_contents_updated_event(self):
        """ Create a business contents updated event. """
        await self._business_object.create_event(
            EventType.BUSINESS_CONTENTS_UPDATED, {}, self._business_object['id']
        )
        logger.info('Business contents updated event created')

    def run(self) -> None:
        """ Run the tasks in the execution pool. """
        async def a_run(_self):
            if self._api_client.playground:
                self._async_pool.ending_tasks['Business_contents_updated'] = (
                    self._create_business_contents_updated_event())
        return self._async_pool.auto_async_func_call(
            name='run',
            func_self=self,
            func=a_run,
        )

    def activate_async_execution(self):
        """ Activate async execution of the tasks. """
        self._async_pool.sequential = False

    def activate_sequential_execution(self):
        """ Activate sequential execution of the tasks. """
        self._async_pool.sequential = True

    def get_api_calls_counter(self):
        """ Get the number of api calls made. """
        return self._api_client.call_counter

    def __getattribute__(self, item):
        """ Get attribute of the client. """
        if False and not IN_BROWSER_PYODIDE:
            if item in ['boards', 'menu_paths'] and not self._business_object:
                log_error(logger, 'Workspace not set. Please use set_workspace() method first.', AttributeError)
            if item in ['activities', 'plt', 'components', 'data', 'io'] and not self._app_object:
                log_error(logger, 'Menu path not set. Please use set_menu_path() method first.', AttributeError)
        return object.__getattribute__(self, item)



import argparse
import subprocess
from abc import ABC
from typing import Optional
import inspect

from aiohttp.client_exceptions import ClientConnectorError

from prompt_toolkit import PromptSession
from prompt_toolkit.completion import Completer, Completion
from prompt_toolkit.buffer import CompletionState
from prompt_toolkit.keys import Keys
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.history import FileHistory

from shimoku.utils import get_args_with_defaults

from shimoku.execution_logger import configure_logging
from shimoku.cli.utils import INTERACTIVE_HISTORY_PATH, get_current_profile

from typing import Iterable
import shlex

import logging

logger = logging.getLogger(__name__)


class CLICustomCompleter(Completer):

    USEFUL_VARIABLES = [
        'environment',
        'access_token',
        'universe_id',
        'workspace_id',
        'local_port',
        'menu_path',
        'board',
        'activity',
    ]
    def __init__(self, parser_dict: dict, variables: dict):
        # parser_dict = {
        #     'name': '',
        #     'commands': [
        #         {...},
        #         ...
        #     ],
        #     'args': [
        #         '--arg_name',
        #         ...
        #     ]
        # }
        self.parser_dict = parser_dict
        self.variables = variables

    def get_completions(self, document, complete_event) -> Iterable[Completion]:
        text = document.text_before_cursor
        if text.startswith('$ '):
            used_vars = [var for var in self.USEFUL_VARIABLES if var.startswith(text[2:]) and var not in self.variables]
            for var in used_vars:
                yield Completion(var, start_position=- len(text[2:]))
            return

        last_char_is_space = text[-1] == ' ' if text else False
        if '  ' in text:
            return []
        words = text.split(' ')

        completion_scope = self.parser_dict
        used_args = []
        next_word_is_arg_value = False
        last_word = '' if not words else words[-1]

        for word in words:
            if next_word_is_arg_value:
                if word in completion_scope['args']:
                    return []
                if last_word == word:
                    break
                next_word_is_arg_value = False
            elif word in [command['name'] for command in completion_scope['commands']]:
                completion_scope = [command for command in completion_scope['commands'] if command['name'] == word][0]
                used_args = []
            elif word in completion_scope['args']:
                next_word_is_arg_value = True
                used_args.append(word)
            elif word and last_char_is_space:
                return []

        if next_word_is_arg_value:
            for var in self.variables:
                if var.startswith(last_word):
                    yield Completion(var, start_position=- len(last_word))
            return []

        for command in completion_scope['commands']:
            if last_char_is_space or command['name'].startswith(last_word):
                yield Completion(command['name'], start_position=- len(last_word))
        for arg in completion_scope['args']:
            if arg not in used_args and (last_char_is_space or arg.startswith(last_word)):
                yield Completion(arg, start_position=- len(last_word))


def create_custom_key_bindings():
    key_bindings = KeyBindings()

    @key_bindings.add(Keys.Tab)
    def _(event):
        """
        Custom Tab key handler: Insert the current completion, trigger completions,
        or auto-complete if only one completion is available.
        """
        buffer = event.app.current_buffer
        current_text = buffer.document.text_before_cursor.rstrip()
        buffer.delete_before_cursor(count=max(len(current_text) - len(current_text.rstrip()), 0))

        completion_state: CompletionState = buffer.complete_state

        if completion_state:
            completions = completion_state.completions
            if completion_state.current_completion:
                # If there's an active completion, apply it and add a space
                buffer.apply_completion(completion_state.current_completion)
                buffer.insert_text(' ')
            elif len(completions) == 1:
                # If there's only one completion, apply it automatically
                buffer.apply_completion(completions[0])
                buffer.insert_text(' ')
            else:
                # Otherwise, move to the next completion
                buffer.complete_next()
        else:
            # If the buffer is empty, start the completion process
            buffer.start_completion()

    return key_bindings


class CLIFuncParam:

    def __init__(
            self,
            name: Optional[str] = None,
            arg_type: Optional[type] = None,
            arg_help: Optional[str] = None,
            mandatory: bool = True,
            default: Optional = None,
            prompt: bool = False,
            choices: Optional[list] = None,
            real_name: Optional[str] = None,
            action: Optional[str] = None,
            group: Optional[str] = None,
            alt_name: Optional[str] = None,
    ):
        self.name = name
        self.arg_type = arg_type
        self.arg_help = arg_help
        self.mandatory = mandatory
        self.default = default
        self.prompt = prompt
        self.choices = choices
        self.real_name = real_name
        self.action = action
        self.group = group
        self.alt_name = alt_name

        if default is not None:
            self.mandatory = False
            if arg_type and not isinstance(default, arg_type):
                raise TypeError(f'The default value {default} is not of type {arg_type}')
            if choices and default not in choices:
                raise ValueError(f'The default value {default} is not in the choices {choices}')

        if prompt or group:
            self.mandatory = False
            if default is not None:
                raise ValueError('Do not set the default value of the CliFuncParam when prompt or group is set to True')


def normalize_name_for_cli(name: str):
    return name.replace('_', '-').lower()


class CLIParser(ABC):

    def __init__(
            self,
            name: str = 'shimoku',
            description: str = "This is the Shimoku cli, it is used to interact with the Shimoku services",
            func: callable = None,
            parent: Optional['CLIParser'] = None,
            arguments: Optional[list[CLIFuncParam]] = None,
    ):
        self.name = name
        self.description = description

        self.command_names: list[str] = []
        self.commands: dict[str, CLIParser] = {}
        self.arguments: list[CLIFuncParam] = []
        self.argument_groups = {}
        self.argument_groups_var_names = {}
        self.shell_commands_enabled = False

        if not parent:
            self.parser = argparse.ArgumentParser(self.name, description=self.description)
        else:
            parent.commands[name] = self
            self.parser = parent.subparsers.add_parser(self.name, description=self.description, help=self.description)

        self.subparsers = self.parser.add_subparsers(dest='command', help='Available commands')

        if arguments:
            for argument in arguments:
                self.add_argument(argument)
        if func:
            self.parser.set_defaults(func=self._decorator_for_self_func(func))
        else:
            self.parser.set_defaults(func=self._default_func)

    def _default_func(self):
        self.parser.print_help()

    def _decorator_for_self_func(self, func: callable):

        def prompt_arg(argument: CLIFuncParam, kwargs: dict):
            if argument.group:
                for arg_name in self.argument_groups_var_names[argument.group]:
                    if kwargs[arg_name] is not None:
                        raise ValueError(f'You can not use the argument {argument.name} with the argument {arg_name}')
            arg_type = argument.arg_type if 'Optional' not in str(argument.arg_type) else argument.arg_type.__args__[0]
            arg_inp = input(f'{argument.real_name if argument.real_name else argument.name}: ')

            kwargs[argument.name] = arg_type(arg_inp) if arg_inp else None

            if argument.choices and kwargs[argument.name] not in argument.choices:
                raise ValueError(f'The value {kwargs[argument.name]} is not in the choices {argument.choices}')

        def wrapper(**kwargs):
            for argument in self.arguments:
                if argument.prompt and kwargs.get(argument.real_name) is None:
                    prompt_arg(argument, kwargs)
                if argument.real_name and argument.name in kwargs:
                    kwargs[argument.real_name] = kwargs.pop(argument.name)

            return func(**kwargs)

        return wrapper

    def decor_add_command(self, name: Optional[str] = None, common_args: list[CLIFuncParam] = None):

        common_args = common_args or []

        def decorator(func: callable):
            func_params = list(inspect.signature(func).parameters.keys())
            func_type_params = {param.name: param.annotation for param in inspect.signature(func).parameters.values()}
            for func_param in func_params:
                if func_param not in func_type_params:
                    func_type_params[func_param] = str
            comment = str(func.__doc__)
            c_help = '\n'.join([line for line in comment.split('\n') if ':param' not in line]).strip()
            params_helps = [line for line in comment.split('\n') if ':param' in line]
            params_helps = [line.split(':param ')[1] for line in params_helps]
            params_helps = {line.split(':')[0].strip(): line.split(':')[1].strip() for line in params_helps}
            default_args = get_args_with_defaults(func)
            func_args = []
            for func_param in func_params:
                if func_param not in params_helps:
                    continue
                dafault_value = default_args.get(func_param)
                if isinstance(dafault_value, CLIFuncParam):
                    func_param_obj = dafault_value
                    if func_param_obj.name is None:
                        func_param_obj.name = normalize_name_for_cli(func_param)
                    if func_param_obj.arg_type is None and func_param_obj.action is None:
                        func_param_obj.arg_type = func_type_params[func_param]
                    if func_param_obj.arg_help is None:
                        func_param_obj.arg_help = params_helps[func_param]
                    if func_param_obj.real_name is None:
                        func_param_obj.real_name = func_param
                    else:
                        raise ValueError('Do not set the real_name of the CliFuncParam, it is set automatically')
                else:
                    arg_type = func_type_params[func_param]
                    mandatory = True
                    if 'Optional' in str(arg_type):
                        arg_type = arg_type.__args__[0]
                        mandatory = False

                    func_param_obj = CLIFuncParam(
                        name=normalize_name_for_cli(func_param),
                        arg_type=arg_type,
                        arg_help=params_helps[func_param],
                        mandatory=mandatory,
                        default=dafault_value,
                    )
                if func_param_obj.name in func_args:
                    raise ValueError(f'The argument {func_param_obj.name} is already registered')
                func_args.append(func_param_obj)

            self.add_command(
                name=normalize_name_for_cli(name if name else func.__name__),
                description=c_help,
                func=func,
                arguments=func_args + common_args,
            )
            return func

        return decorator

    def add_command(
            self,
            name: str = None,
            description: str = None,
            func: callable = None,
            arguments: Optional[list[CLIFuncParam]] = None
    ) -> 'CLIParser':
        command = CLIParser(
            name=name, description=description,
            func=func, parent=self, arguments=arguments
        )
        if command.name in self.command_names:
            raise ValueError(f'The command {command.name} is already registered')
        self.command_names.append(command.name)
        return command

    def add_argument(self, argument: CLIFuncParam):
        def add_prefix(name: str):
            return ('-' if len(name) == 1 else '--') + name

        if argument.name in self.arguments:
            raise ValueError(f'The argument {argument.name} is already registered')
        params = {
            k: v for k, v in dict(action=argument.action, type=argument.arg_type, help=argument.arg_help,
                                  required=argument.mandatory, default=argument.default,
                                  choices=argument.choices).items()
            if v is not None
        }
        argument_names = [add_prefix(argument.name)]
        if argument.alt_name:
            argument_names.append(add_prefix(argument.alt_name))

        if argument.group:
            if argument.group not in self.argument_groups:
                self.argument_groups[argument.group] = self.parser.add_mutually_exclusive_group()
                self.argument_groups_var_names[argument.group] = []
            self.argument_groups[argument.group].add_argument(*argument_names, **params)
            self.argument_groups_var_names[argument.group].append(argument.name)
        else:
            self.parser.add_argument(*argument_names, **params)
        self.arguments.append(argument)

    @staticmethod
    async def execute_args(args, user_vars: Optional[dict] = None):
        args_ = vars(args).copy()
        args_.pop('interactive', None)
        args_.pop('shell_commands_enabled', None)
        args_.pop('command', None)
        args_.pop('func', None)

        if user_vars:
            user_vars = {k.replace('$', '').replace('-', '_'): v for k, v in user_vars.items()}
            for func_param in args_.keys():
                if args_[func_param] is None and func_param in user_vars:
                    args_[func_param] = user_vars.get(func_param)
                    logger.info(f"Using saved value for unspecified argument {func_param}")

        result = args.func(**args_)
        if inspect.isawaitable(result):
            await result

    async def parse_args(self):
        # argcomplete.autocomplete(self.parser)
        try:
            args = self.parser.parse_args()
        except SystemExit:
            return
        if hasattr(args, 'interactive') and args.interactive:
            self.shell_commands_enabled = args.shell_commands_enabled
            await self.parse_args_interactive()
            return
        await self.execute_args(args)

    def gather_commands_and_options(self):
        commands_and_options = {'name': self.name, 'commands': [], 'args': []}
        for command in self.commands.values():
            commands_and_options['commands'].append(command.gather_commands_and_options())
        for argument in self.arguments:
            if argument.name in ['interactive', 'shell-commands-enabled']:
                continue
            commands_and_options['args'].append('--' + argument.name)
        commands_and_options['args'].extend(['--help', '-h'])
        return commands_and_options

    async def parse_args_interactive(self):
        # Dictionary to store user-defined variables
        variables = {}

        # Gather possible commands and options
        commands_and_options = self.gather_commands_and_options()
        completer = CLICustomCompleter(commands_and_options, variables)

        # Trim the history file to 1000 lines
        with open(INTERACTIVE_HISTORY_PATH, 'r+') as file:
            lines = file.readlines()
            trimmed_lines = lines[-1000:]
            file.seek(0)
            file.truncate()
            file.writelines(trimmed_lines)

        session = PromptSession(
            completer=completer,
            key_bindings=create_custom_key_bindings(),
            history=FileHistory(INTERACTIVE_HISTORY_PATH)
        )
        configure_logging('INFO')

        while True:
            try:
                text = await session.prompt_async(f'S-{get_current_profile().upper()}> ')
                if not text:
                    continue

                # Check for variable assignment
                if text.startswith('$'):
                    try:
                        print()
                        # Splitting the input while respecting quoted sections
                        parts = shlex.split(text)
                        for i in range(1, len(parts), 2):
                            var_name, var_value = parts[i], parts[i + 1]
                            variables['$' + var_name] = var_value
                            print(f"Variable {var_name} set to {var_value}")
                            print()
                    except Exception as e:
                        if isinstance(e, (ValueError, IndexError)):
                            print("Invalid variable assignment syntax.")
                            print()
                        else:
                            raise e
                    continue
                if text.startswith('unset$'):
                    _, *var_names = text.split(' ')
                    for var_name in var_names:
                        if '$' + var_name not in variables:
                            print(f"Variable {var_name} not set")
                            print()
                            continue
                        variables.pop('$' + var_name)
                        print(f"Variable {var_name} unset")
                        print()
                    continue

                # Replace variables in user input
                for var_name, var_value in variables.items():
                    text = text.replace(var_name, var_value)

                if '$' in text:
                    print("Warning: It's possible that not all variables were replaced, please check your input.")
                    print()
                if text in ['exit', 'esc']:
                    break
            except KeyboardInterrupt:
                break
            except EOFError:
                break

            if text.startswith('show '):
                text = text.replace('show ', '')
                print(text)
                print()
                continue

            try:
                args = self.parser.parse_args(text.rstrip().split(' '))
                await self.execute_args(args, variables)
            except SystemExit as e:
                if not self.shell_commands_enabled or e.code == 0:
                    continue
                print()
                confirmation = input(f'The command: "{text}" \n'
                                     f'could not be interpreted by the Shimoku parser\n'
                                     'do you want to execute it as a shell command? (y/n): ')
                if confirmation.lower() == 'y':
                    print()
                    try:
                        subprocess.run(text.split(), check=True)
                    except FileNotFoundError:
                        print(f"Command not found: {text}")
                    except subprocess.CalledProcessError:
                        print(f"Command failed: {text}")
            except Exception as e:
                if isinstance(e,
                              (ValueError, TypeError, KeyError,
                               AttributeError, RuntimeError, IndexError,
                               ClientConnectorError)
                              ):
                    logger.error(e)



import uuid
import ast
from typing import Dict, Optional
from shimoku.utils import IN_BROWSER_PYODIDE
from shimoku.exceptions import ActionError
from shimoku.execution_logger import log_error

import logging

logger = logging.getLogger(__name__)


class Scope:
    def __init__(self, parent_scope: Optional['Scope'] = None):
        self.lines_need_async = set()
        self.inner_scopes: Dict[str, Scope] = {}
        self.needs_async = False
        self.parent_scope: Optional[Scope] = parent_scope
        self.parameters: dict = {}

    def get_function(self, name: str):
        if name in self.inner_scopes:
            return self.inner_scopes[name]
        if self.parent_scope is not None:
            return self.parent_scope.get_function(name)
        return None

    def user_defined_scope(self):
        return self.parent_scope is not None and self.parent_scope.parent_scope is not None

    def enter_scope(self, name: str):
        if name not in self.inner_scopes:
            self.inner_scopes[name] = Scope(self)
        return self.inner_scopes[name]

    def exit_scope(self):
        return self.parent_scope


class ActionsChecker(ast.NodeVisitor):

    shimoku_client_attrs = [
        'universes', 'workspaces', 'boards', 'menu_paths',
        'components', 'data', 'io', 'activities', 'plt',
        'set_workspace', 'set_board', 'set_menu_path', 'run'
    ]

    SHIMOKU_CLIENT_NAME = 'shimoku_client'

    def __init__(self):
        self.current_scope = Scope()
        self.in_class = False
        self.exists_action_function = False
        self.parameters = {}
        self.client_assigned = False
        self.shimoku_imported = False

    def is_shimoku_client(self, node):
        if isinstance(node, ast.Name) and node.id == self.SHIMOKU_CLIENT_NAME:
            return True
        if (isinstance(node, ast.Call) and
                isinstance(node.func, ast.Name) and
                node.func.id == 'Client'):
            return True
        return False

    def visit_Import(self, node):
        if not self.current_scope.parent_scope:
            return
        for alias in node.names:
            if 'asyncio' in alias.name:
                log_error(logger, f"Cannot import asyncio module.", ActionError)

    def visit_ImportFrom(self, node):
        if node.module == 'shimoku':
            if any(asname is not None for asname in [alias.asname for alias in node.names]):
                log_error(logger, f"Cannot import shimoku modules with renaming.", ActionError)
            for alias in node.names:
                if alias.name == 'Client':
                    self.shimoku_imported = True
        if not self.current_scope.parent_scope:
            return
        if node.module == 'asyncio':
            log_error(logger, f"Cannot import from asyncio module.", ActionError)

    def visit_Assign(self, node):
        for target in node.targets:
            if self.is_shimoku_client(node.value):
                if self.client_assigned:
                    log_error(logger, f"Cannot assign Client instance", ActionError)
                self.client_assigned = True
            elif isinstance(target, ast.Name):
                if target.id == self.SHIMOKU_CLIENT_NAME:
                    log_error(logger, f"Cannot assign to {self.SHIMOKU_CLIENT_NAME} variable.", ActionError)

        self.generic_visit(node)

    def visit_FunctionDef(self, node):
        if node.name == 'action':
            if self.in_class:
                log_error(logger, f"Cannot define action function inside a class.", ActionError)
            if self.current_scope.user_defined_scope():
                log_error(logger, f"Cannot define action function inside another function.", ActionError)
            if self.exists_action_function:
                log_error(logger, f"Cannot define more than one action function.", ActionError)
            if len(node.args.args) != 1:
                log_error(logger, f"Action function must have exactly one parameter.", ActionError)
            if node.args.args[0].arg != self.SHIMOKU_CLIENT_NAME:
                log_error(logger, f"Action function parameter must be named 'shimoku_client'.", ActionError)
            if node.args.args[0].annotation and node.args.args[0].annotation.id != 'Client':
                log_error(logger, f"Action function parameter must be annotated with 'Client'.", ActionError)
            self.exists_action_function = True

        self.current_scope = self.current_scope.enter_scope(node.name)
        self.generic_visit(node)

        for arg in node.args.args:
            if self.in_class and arg.arg == 'self':
                continue

            if not arg.annotation:
                log_error(logger, f"Annotations are mandatory for function parameters.", ActionError)
            if isinstance(arg.annotation, ast.Subscript):
                if 'Client' in ast.dump(arg.annotation):
                    log_error(
                        logger,
                        f"Cannot use Client instance as a "
                        f"part of a parameter type, use 'Client' instead.",
                        ActionError
                    )
            elif arg.annotation.id == 'Client':
                if arg.arg != self.SHIMOKU_CLIENT_NAME:
                    log_error(
                        logger,
                        f"Client parameter must be named '{self.SHIMOKU_CLIENT_NAME}' not '{arg.arg}'.",
                        ActionError
                    )
            elif arg.arg == self.SHIMOKU_CLIENT_NAME:
                log_error(
                    logger,
                    f"{self.SHIMOKU_CLIENT_NAME} parameter must be annotated with "
                    f"'Client' not '{arg.annotation.id}'.",
                    ActionError
                )
            self.current_scope.parameters[arg.arg] = arg.annotation.id

        self.current_scope = self.current_scope.exit_scope()

    def visit_AsyncFunctionDef(self, node):
        if not self.current_scope.parent_scope:
            self.current_scope = self.current_scope.enter_scope(node.name)
            self.generic_visit(node)
            self.current_scope = self.current_scope.exit_scope()
        else:
            log_error(logger, f"Cannot use async functions.", ActionError)

    def visit_ClassDef(self, node):
        self.in_class = True
        self.generic_visit(node)
        self.in_class = False

    def visit_Return(self, node):
        if isinstance(node.value, ast.Name):
            if node.value.id == self.SHIMOKU_CLIENT_NAME:
                log_error(logger, f"Cannot return Client instance '{node.value.id}'.", ActionError)
        self.generic_visit(node)

    def visit_Call(self, node):
        self.generic_visit(node)

        # Check if the call is a method call that requires await
        if isinstance(node.func, ast.Attribute):
            aux_node = node.func
            while not isinstance(aux_node.value, ast.Name):
                if not isinstance(aux_node.value, ast.Attribute):
                    return
                aux_node = aux_node.value
            if aux_node.value.id == self.SHIMOKU_CLIENT_NAME and aux_node.attr in self.shimoku_client_attrs:
                print(f"Client method call '{node.func.attr}' used at line {node.lineno} without await.")
                self.current_scope.lines_need_async.add((node.lineno, node.col_offset))
                self.current_scope.needs_async = True

        if not isinstance(node.func, ast.Name) or self.current_scope.get_function(node.func.id) is None:
            return

        function_scope = self.current_scope.get_function(node.func.id)
        if function_scope.needs_async:
            self.current_scope.needs_async = True

        parameters = function_scope.parameters
        parameter_types = list(parameters.values())
        for i, arg in enumerate(node.args):
            if self.is_shimoku_client(arg) and not parameter_types[i] == 'Client':
                log_error(
                    logger,
                    f"The value asigned to the parameter '{arg.id}' is the Client, "
                    f"but the function '{node.func.id}' expects a '{parameter_types[i]}' instance.",
                    ActionError
                )
        for keyword in node.keywords:
            if self.is_shimoku_client(keyword.value) and not parameters[keyword.arg] == 'Client':
                log_error(
                    logger,
                    f"The value asigned to the parameter '{keyword.arg}' is the Client, "
                    f"but the function '{node.func.id}' expects a '{parameters[keyword.arg]}' instance.",
                    ActionError
                )


class AwaitInserter(ast.NodeTransformer):

    def __init__(self, scope: Scope):
        self.current_scope = scope

    def visit_Call(self, node):
        self.generic_visit(node)
        if not self.current_scope.parent_scope:
            return node

        if (node.lineno, node.col_offset) in self.current_scope.lines_need_async:
            self.current_scope.needs_async = True
            return ast.copy_location(ast.Await(value=node), node)

        # Check if the call is a function call that requires await
        func_name = node.func.id if isinstance(node.func, ast.Name) else node.func.attr
        function_scope = self.current_scope.get_function(func_name)
        if function_scope is not None and function_scope.needs_async:
            self.current_scope.needs_async = True
            print(f"Function call '{func_name}' used at line {node.lineno} without await.")
            return ast.copy_location(ast.Await(value=node), node)

        return node

    def visit_FunctionDef(self, node):
        self.current_scope = self.current_scope.enter_scope(node.name)
        node = self.generic_visit(node)
        async_node = None
        if self.current_scope.needs_async:
            # Convert to an async function
            async_node = ast.AsyncFunctionDef(
                name=node.name,
                args=node.args,
                body=node.body,
                decorator_list=node.decorator_list,
                returns=node.returns,
                type_comment=node.type_comment
            )
        self.current_scope = self.current_scope.exit_scope()
        if async_node is not None:
            return ast.copy_location(async_node, node)
        return node

    def visit_AsyncFunctionDef(self, node):
        self.current_scope = self.current_scope.enter_scope(node.name)
        node = self.generic_visit(node)
        self.current_scope = self.current_scope.exit_scope()
        return node


def execute_action(code: str, print_code: bool = False):
    code = code.replace('\r\n', '\n').replace('\\\n', '').replace('\\\r', '')
    code_lines = code.split('\n')
    main_uuid = str(uuid.uuid4()).replace('-', '_')

    script_code_lines = [
        'import asyncio',
        '',
        f'async def main_{main_uuid}():',
        *[f'    {line}' for line in code_lines],
        '    shimoku_client = Client(',
        '        async_execution=True,',
        '        verbosity="INFO",',
        '    )',
        '    action(shimoku_client)',
        '',
    ]
    if IN_BROWSER_PYODIDE:
        script_code_lines.append(f'asyncio.get_event_loop().create_task(main_{main_uuid}())', )
    else:
        script_code_lines.append(f'asyncio.run(main_{main_uuid}())', )

    if print_code:
        for i, line in enumerate(script_code_lines):
            print(f'{i:3} | {line}')

    script_ast = ast.parse('\n'.join(script_code_lines))

    analyzer = ActionsChecker()
    analyzer.visit(script_ast)
    if not analyzer.exists_action_function:
        log_error(logger, f"Action function is not defined.", ActionError)
    if not analyzer.shimoku_imported:
        log_error(logger, f"Shimoku's Client is not imported.", ActionError)
    modifier = AwaitInserter(analyzer.current_scope)
    modifier.visit(script_ast)
    dump = ast.dump(script_ast, indent=4).split('\n')
    # for i, line in enumerate(dump):
    #     print(f'{i:3} | {line}')
    exec(compile(script_ast, filename='shimoku_action', mode='exec'))


def main():
    # with open('../../../tests/unit_tests/test_plot_api.py', 'r') as f:
    with open('../../../tests/personal_test.py', 'r') as f:
        code_lines = f.readlines()
    execute_action('\n'.join(code_lines))


if __name__ == '__main__':
    main()



from typing import TYPE_CHECKING
from copy import deepcopy
from shimoku.api.utils import revert_uuids_from_dict, change_data_set_name_with_report
from shimoku.code_gen.business_code_gen.apps_code_gen. \
    data_sets_code_gen.code_gen_from_data_sets import code_gen_read_csv_from_data_set, get_linked_data_set_info
from shimoku.code_gen.utils_code_gen import code_gen_from_dict, code_gen_from_list
if TYPE_CHECKING:
    from shimoku.api.resources.report import Report
    from shimoku.code_gen.business_code_gen.apps_code_gen.code_gen_from_apps import AppCodeGen

import logging
from shimoku.execution_logger import log_error
logger = logging.getLogger(__name__)


def get_fields_from_mapping(rev_columns, mappings, data_is_custom):
    if data_is_custom:
        return ['data']
    fields = []
    for _, mapping in mappings:
        if isinstance(mapping, dict):
            entry = {}
            for k, v in mapping.items():
                entry[k] = rev_columns[v] if v in rev_columns else v
            fields.append(entry)
        elif isinstance(mapping, list):
            entry = []
            for v in mapping:
                entry.append(rev_columns[v] if v in rev_columns else v)
            fields.append(entry)
        elif mapping in rev_columns:
            fields.append(rev_columns[mapping])
        else:
            fields.append(mapping)
    return fields


async def code_gen_data_arg(self: 'AppCodeGen', report, data_set_id, data_set, data_is_custom):
    data_arg_code_lines = ['[{}],']

    if data_set_id in self._code_gen_tree.shared_data_sets:
        data_arg_code_lines = [f'    "{data_set["name"]}",']
        if data_is_custom:
            data_arg_code_lines += ['    data_is_not_df=True,']
    elif data_is_custom:
        val = self._code_gen_tree.custom_data_sets_with_data[data_set_id]
        data_arg_code_lines = code_gen_from_dict(val, 4) \
            if isinstance(val, dict) else code_gen_from_list(val, 4)
        data_arg_code_lines += ['    data_is_not_df=True,']
    elif data_set is not None:
        data_arg_code_lines = [
            '    '+(await code_gen_read_csv_from_data_set(
                data_set, change_data_set_name_with_report(data_set, report)
            ))
        ]
        if data_arg_code_lines[0] is None:
            return ['pass']
        data_arg_code_lines[0] += ','

    return data_arg_code_lines


async def code_gen_from_echarts(
        self: 'AppCodeGen', report: 'Report', properties: dict
) -> list[str]:
    """ Generate code for an echarts report.
    :param report: report to generate code from
    :param report_params: parameters of the report
    :param properties: properties of the report
    :return: list of code lines
    """
    echart_options = deepcopy(properties['option'])
    rds_ids_in_order = revert_uuids_from_dict(echart_options)
    referenced_data_sets, mappings = await get_linked_data_set_info(self, report, rds_ids_in_order)
    if len(referenced_data_sets) > 1:
        log_error(logger,
                  'Only one data set is supported for the current implementation of the echarts component.',
                  RuntimeError)

    data_set_id, data_set = list(referenced_data_sets.items())[0] if len(referenced_data_sets) > 0 else (None, None)
    rev_columns = {}
    if data_set_id is not None and data_set['columns']:
        rev_columns = {v: k for k, v in data_set['columns'].items()}

    data_is_custom = data_set_id in self._code_gen_tree.custom_data_sets_with_data
    fields = code_gen_from_list(get_fields_from_mapping(rev_columns, mappings, data_is_custom), 4)
    data_arg = await code_gen_data_arg(self, report, data_set_id, data_set, data_is_custom)

    options_code = code_gen_from_dict(echart_options, 4)

    return [
        'shimoku_client.plt.free_echarts(',
        *self.code_gen_report_params(report.cascade_to_dict()),
        f'    data={data_arg[0]}',
        *data_arg[1:],
        f'    fields={fields[0][4:]}',
        *fields[1:],
        f'    options={options_code[0][4:]}',
        *options_code[1:],
        ')'
    ]



from typing import Optional, Union, TYPE_CHECKING

from pandas import DataFrame

from shimoku.plt.EChart_definitions.default_echart_options import get_common_echart_options, get_common_series_options

if TYPE_CHECKING:
    from shimoku.plt.plt_layer import PlotLayer


async def radar_chart(
    self: 'PlotLayer', order: int, names: str, data: Union[str, DataFrame, list[dict]],
    values: Optional[list[str]] = None, max_field: Optional[str] = None, fill_area: bool = False,
    title: Optional[str] = None, rows_size: Optional[int] = None,
    cols_size: Optional[int] = None, padding: Optional[str] = None,
    option_modifications: Optional[dict] = None,
):
    """ Create a radar chart """
    common_options = get_common_echart_options()
    series_options = get_common_series_options()
    series_options['type'] = 'radar'
    del common_options['yAxis']
    del common_options['xAxis']
    del common_options['grid']
    del common_options['toolbox']['feature']['magicType']
    common_options['radar'] = {'indicator': '#set_data#'}
    if fill_area:
        series_options['areaStyle'] = {'opacity': 0.25}

    data_mapping_to_tuples = await self._choose_data(order, data)
    if not values:
        values = [f for f in data_mapping_to_tuples.keys() if f != names and f != max_field]

    series_options['data'] = [{'value': '#set_data#', 'name': vf} for vf in values]
    common_options['series'] = [series_options]
    indicator_fields = {'name': names} if not max_field else {'name': names, 'max': max_field}
    await self._create_echart(
        fields=[indicator_fields, *values], options=common_options,
        data_mapping_to_tuples=data_mapping_to_tuples, option_modifications=option_modifications,
        order=order, title=title, rows_size=rows_size, cols_size=cols_size, padding=padding)



from os import getenv
import unittest
from utils import initiate_shimoku

s = initiate_shimoku()

business_id: str = getenv('BUSINESS_ID')
s.set_workspace(uuid=business_id)


class TestRoles(unittest.TestCase):

    def test_business_roles(self):

        role = s.workspaces.get_role(uuid=business_id, role_name='test_role')
        if role:
            s.workspaces.delete_role(uuid=business_id, role_id=role['id'])

        previous_roles_size = len(s.workspaces.get_roles(uuid=business_id))
        role = s.workspaces.create_role(uuid=business_id, role_name='test_role')
        assert role['role'] == 'test_role'
        assert role['permission'] == 'READ'
        assert role['resource'] == 'BUSINESS_INFO'
        assert role['target'] == 'GROUP'

        roles = s.workspaces.get_roles(uuid=business_id)

        assert len(roles) == previous_roles_size + 1

        s.workspaces.delete_role(uuid=business_id, role_id=role['id'])

        assert len(s.workspaces.get_roles(uuid=business_id)) == previous_roles_size

        s.workspaces.create_role(uuid=business_id, role_name='test_role', permission='WRITE',
                                 resource='DATA', target='USER')

        role = [role_ for role_ in s.workspaces.get_roles(uuid=business_id) if role_['role'] == 'test_role'][0]

        assert role['permission'] == 'WRITE'
        assert role['resource'] == 'DATA'
        assert role['target'] == 'USER'

        assert len(s.workspaces.get_roles(uuid=business_id)) == previous_roles_size + 1

        s.workspaces.delete_role(uuid=business_id, role_id=role['id'])

    def test_dashboard_roles(self):
        dashboard_name = 'roles_dashboard'
        dashboard = s.boards.get_board(name=dashboard_name)
        if not dashboard:
            dashboard = s.boards.create_board(name=dashboard_name)

        role = s.boards.get_role(name=dashboard_name, role_name='test_role')
        if role:
            s.boards.delete_role(name=dashboard_name, role_id=role['id'])

        role = s.boards.create_role(name=dashboard_name, role_name='test_role')
        assert role['role'] == 'test_role'
        assert role['permission'] == 'READ'
        assert role['resource'] == 'BUSINESS_INFO'
        assert role['target'] == 'GROUP'

        roles = s.boards.get_roles(name=dashboard_name)

        assert len(roles) == 1
        assert roles[0]['role'] == 'test_role'

        s.boards.delete_role(name=dashboard_name, role_id=role['id'])

        assert len(s.boards.get_roles(name=dashboard_name)) == 0

        s.boards.create_role(name=dashboard_name, role_name='test_role', permission='WRITE',
                             resource='DATA', target='USER')

        role = s.boards.get_roles(name=dashboard_name)[0]

        assert role['role'] == 'test_role'
        assert role['permission'] == 'WRITE'
        assert role['resource'] == 'DATA'
        assert role['target'] == 'USER'

        assert len(s.boards.get_roles(name=dashboard_name)) == 1

        s.boards.delete_board(uuid=dashboard['id'])

    def test_app_roles(self):
        app = s.menu_paths.get_menu_path(name='roles_app')

        role = s.menu_paths.get_role(uuid=app['id'], role_name='test_role')
        if role:
            s.menu_paths.delete_role(uuid=app['id'], role_id=role['id'])

        role = s.menu_paths.create_role(uuid=app['id'], role_name='test_role')
        assert role['role'] == 'test_role'
        assert role['permission'] == 'READ'
        assert role['resource'] == 'BUSINESS_INFO'
        assert role['target'] == 'GROUP'

        roles = s.menu_paths.get_roles(uuid=app['id'])

        assert len(roles) == 1
        assert roles[0]['role'] == 'test_role'

        s.menu_paths.delete_role(uuid=app['id'], role_id=role['id'])

        assert len(s.menu_paths.get_roles(uuid=app['id'])) == 0

        s.menu_paths.create_role(uuid=app['id'], role_name='test_role', permission='WRITE', resource='DATA', target='USER')

        role = s.menu_paths.get_roles(uuid=app['id'])[0]

        assert role['role'] == 'test_role'
        assert role['permission'] == 'WRITE'
        assert role['resource'] == 'DATA'
        assert role['target'] == 'USER'

        assert len(s.menu_paths.get_roles(uuid=app['id'])) == 1

        s.menu_paths.delete_menu_path(uuid=app['id'])



from shimoku.cli.configuration import add_configuration_parser
from shimoku.cli.cloud_parser import add_cloud_parser
from shimoku.cli.playground import add_playground_parser
from shimoku.cli.persistence import add_persistence_parser
from shimoku.cli import CLIParser, CLIFuncParam

from shimoku.execution_logger import configure_logging

import asyncio

main_parser = CLIParser(
    arguments=[
        CLIFuncParam(
            name='interactive',
            arg_help='Run in interactive mode',
            mandatory=False,
            default=False,
            action='store_true',
            alt_name='i'
        ),
        CLIFuncParam(
            name='shell-commands-enabled',
            arg_help='Enable the possibility to run shell commands from the interactive mode',
            mandatory=False,
            default=False,
            action='store_true',
            alt_name='s'
        ),
    ]
)

add_configuration_parser(main_parser)
add_playground_parser(main_parser)
add_cloud_parser(main_parser)
add_persistence_parser(main_parser)


def main():
    configure_logging('INFO')
    asyncio.run(main_parser.parse_args())


if __name__ == "__main__":
    main()



from typing import Optional, Union, TYPE_CHECKING

from shimoku.plt.utils import validate_data_is_pandarable, ShimokuPalette
from shimoku.exceptions import DataError

from pandas import DataFrame

from shimoku.plt.EChart_definitions.default_echart_options import get_common_echart_options

if TYPE_CHECKING:
    from shimoku.plt.plt_layer import PlotLayer

import logging
from shimoku.execution_logger import log_error
logger = logging.getLogger(__name__)


async def waterfall_chart(
    self: 'PlotLayer',
    data: Union[str, DataFrame, list[dict]], order: int,
    x: str, positive: str, negative: str,
    rows_size: Optional[int] = None, cols_size: Optional[int] = None,
    padding: Optional[str] = None, title: Optional[str] = None,
    x_axis_name: Optional[str] = None, y_axis_name: Optional[str] = None,
    show_balance: bool = False, variant: Optional[str] = None,
    option_modifications: Optional[dict] = None,
):
    if isinstance(data, str):
        log_error(logger, f'Waterfall cannot use shared data, please provide a DataFrame or a list of dicts', DataError)

    df: DataFrame = validate_data_is_pandarable(data)

    df['Base'] = (df[positive].cumsum() - df[negative].cumsum()).shift(1).fillna(0)
    df['Value'] = df[positive] - df[negative]
    df['Balance'] = df['Base'] + df['Value']
    df['VisibleNegBase'] = 0
    df['VisiblePosBase'] = 0

    df['Positive'] = df['Value'].apply(lambda x: x if x > 0 else 0)
    df['Negative'] = df['Value'].apply(lambda x: abs(x) if x < 0 else 0)

    def change_sign(row):
        if row['Base'] < 0:
            if row['Base'] + row['Positive'] > 0:
                row['VisiblePosBase'] = -(row['Positive'] - (row['Base'] + row['Positive']))
                row['Positive'] = row['Base'] + row['Positive']
                row['Base'] = 0
                return row
            else:
                row['Positive'] = -row['Positive']
                row['Negative'] = -row['Negative']
        else:
            if row['Base'] - row['Negative'] < 0:
                row['VisibleNegBase'] = row['Negative'] + (row['Base'] - row['Negative'])
                row['Negative'] = row['Base'] - row['Negative']
                row['Base'] = 0
                return row

        row['Base'] = row['Base'] - row['Positive'] if row['Base'] < 0 else row['Base'] - row['Negative']
        return row

    df = df.apply(change_sign, axis=1)
    df = df[[x, 'Balance', 'Base', 'VisibleNegBase', 'VisiblePosBase', 'Negative', 'Positive']].fillna(0)

    data_mappings_to_tuples = await self._choose_data(order, df)

    common_options = get_common_echart_options()
    del common_options['toolbox']['feature']['magicType']
    common_options.update({
        'legend': {
            'data': [positive, negative] + (['Balance'] if show_balance else []),
            'show': True,
            'type': 'scroll',
            'itemGap': 16,
            'icon': 'circle'
        },
        'xAxis': [
            {
                'data': '#set_data#',
                'type': 'category',
                'fontFamily': 'Rubik',
                'nameLocation': 'middle',
                'nameGap': 35,
            }
        ],
        'yAxis': [
            {
                'type': 'value',
                'fontFamily': 'Rubik',
                'nameLocation': 'middle',
                'nameGap': int(24+7*(df['Balance'].astype(str).apply(len).max())),
            },
        ],
    })
    series_options = [
        {
            'data': '#set_data#',
            'name': 'Balance',
            'step': 'end',
            'symbol': 'none',
            'type': 'line',
            'itemStyle': {
                'color': ShimokuPalette.CHART_C3.value if show_balance else 'transparent',
            },
            'emphasis': {
                'itemStyle': {
                    'color': ShimokuPalette.CHART_C3.value if show_balance else 'transparent',
                },
                'lineStyle': {
                    'color': ShimokuPalette.CHART_C3.value if show_balance else 'transparent',
                }
            },
            'zlevel': 0,
            'tooltip': {'show': show_balance},
        },
        {
            'data': '#set_data#',
            'name': 'Placeholder',
            'type': 'bar',
            'stack': 'Total',
            'itemStyle': {
                'borderColor': 'transparent',
                'color': 'transparent'
            },
            'emphasis': {
                'itemStyle': {
                    'borderColor': 'transparent',
                    'color': 'transparent'
                }
            },
            'tooltip': {'show': False},
        },
        {
            'data': '#set_data#',
            'name': 'Negative visible Base',
            'type': 'bar',
            'stack': 'Total',
            'silent': True,
            'itemStyle': {
                'color': ShimokuPalette.CHART_C6.value,
            },
            'emphasis': {
                'itemStyle': {
                    'color': ShimokuPalette.CHART_C6.value,
                },
            },
            'zlevel': 1,
            'tooltip': {'show': False},
        },
        {
            'data': '#set_data#',
            'name': 'Positive visible Base',
            'type': 'bar',
            'stack': 'Total',
            'silent': True,
            'itemStyle': {
                'color': ShimokuPalette.CHART_C2.value,
            },
            'emphasis': {
                'itemStyle': {
                    'color': ShimokuPalette.CHART_C2.value,
                },
            },
            'zlevel': 1,
            'tooltip': {'show': False},
        },
        {
            'data': '#set_data#',
            'name': negative,
            'type': 'bar',
            'stack': 'Total',
            'itemStyle': {
                'borderRadius': [0, 0, 9, 9],
                'color': ShimokuPalette.CHART_C6.value,
            },
            'emphasis': {
                'itemStyle': {
                    'borderRadius': [0, 0, 9, 9],
                    'color': ShimokuPalette.CHART_C6.value,
                },
            },
            'zlevel': 1,
        },
        {
            'data': '#set_data#',
            'name': positive,
            'type': 'bar',
            'stack': 'Total',
            'itemStyle': {
                'borderRadius': [9, 9, 0, 0],
                'color': ShimokuPalette.CHART_C2.value,
            },
            'emphasis': {
                'itemStyle': {
                    'borderRadius': [9, 9, 0, 0],
                    'color': ShimokuPalette.CHART_C2.value,
                },
            },
            'zlevel': 1,
        },
    ]

    await self._create_trend_chart(
        data_mapping_to_tuples=data_mappings_to_tuples, order=order, axes=x,
        values=['Balance', 'Base', 'VisibleNegBase', 'VisiblePosBase', 'Negative', 'Positive'],
        rows_size=rows_size, cols_size=cols_size, padding=padding, title=title, x_axis_names=x_axis_name,
        y_axis_names=y_axis_name, echart_options=common_options, series_options=series_options,
        option_modifications=option_modifications, variant=variant
    )



from typing import Optional, Union, TYPE_CHECKING

from shimoku.plt.EChart_definitions.default_echart_options import get_common_echart_options, get_common_series_options

if TYPE_CHECKING:
    from shimoku.plt.plt_layer import PlotLayer

import logging
logger = logging.getLogger(__name__)


async def tree_chart(
    self: 'PlotLayer', order: int, data: Union[str, dict, list[dict]],
    radial: bool = False, vertical: bool = False,
    title: Optional[str] = None, padding: Optional[str] = None,
    rows_size: Optional[int] = None, cols_size: Optional[int] = None,
    option_modifications: Optional[dict] = None,
):
    """ Create a tree chart """
    if radial and vertical:
        logger.warning('Radial and vertical options are mutually exclusive. Vertical option will be ignored.')
    common_options = get_common_echart_options()
    series_options = get_common_series_options()
    del common_options['xAxis']
    del common_options['yAxis']
    del common_options['toolbox']['feature']['magicType']
    series_options.update({
        'type': 'tree',
        'data': '#set_data#',
        'symbolSize': 7,
        'initialTreeDepth': 3,
        'layout': 'orthogonal' if not radial else 'radial',
        'orient': 'horizontal' if not vertical else 'vertical',
        'label': {
            'position': 'left' if not vertical else 'top',
            'rotate': 0 if not vertical else -90,
            'verticalAlign': 'middle',
            'align': 'right',
            'fontSize': 9
        },
        'leaves': {
            'label': {
                'position': 'right' if not vertical else 'bottom',
                'rotate': 0 if not vertical else -90,
                'verticalAlign': 'middle',
                'align': 'left' if not vertical else 'right'
            }
        },
        'emphasis': {
            'focus': 'descendant'
        },
        'expandAndCollapse': True,
        'animationDuration': 550,
        'animationDurationUpdate': 750
    })
    common_options.update({
        'tooltip': {
            'trigger': 'item',
            'triggerOn': 'mousemove'
        },
        'series': series_options
    })
    if radial:
        del common_options['series']['orient']
        del common_options['series']['leaves']
        del common_options['series']['label']

    if isinstance(data, dict):
        data = [data]

    await self._create_echart(
        data_mapping_to_tuples=await self._choose_data(order, data, dump_whole=True),
        fields=['data'], options=common_options, order=order, title=title, option_modifications=option_modifications,
        rows_size=rows_size, cols_size=cols_size, padding=padding)



""""""
from os import getenv
from typing import Dict, List
import unittest

from shimoku.exceptions import MenuPathError, APIError
from utils import initiate_shimoku

s = initiate_shimoku()
mock: bool = getenv('MOCK') == 'TRUE'
business_id: str = getenv('BUSINESS_ID')
s.set_workspace(uuid=business_id)

app_name = 'Main test app'
s.set_menu_path(app_name, 'test')
s.plt.html(html='<h1>test</h1>', order=0)


# Check it does not exists anymore
class TestApp(unittest.TestCase):
    def check_app_not_exists(self, _app_id: str):
        if mock:
            return
        with self.assertRaises(APIError):
            s.menu_paths.get_menu_path(uuid=_app_id)

    def test_get_app(self):
        app_id = s.menu_paths.get_menu_path(name=app_name)['id']
        app: Dict = s.menu_paths.get_menu_path(uuid=app_id)
        assert app

    def test_get_fake_app(self):
        if not mock:
            with self.assertRaises(APIError):
                app_id: str = 'this is a test'
                s.menu_paths.get_menu_path(uuid=app_id)

    def test_update_app(self):
        """Set the updatedAt field of an App to '2000-01-01'
        Then revert the updatedAt to its original value
        """
        app: Dict = s.menu_paths.get_menu_path(name=app_name)
        app_id: str = app['id']
        old_vals = dict(
            hide_title=app['hideTitle'],
            hide_path=app['hidePath'],
            show_breadcrumb=app['showBreadcrumb'],
            show_history_navigation=app['showHistoryNavigation'],
        )

        new_vals = dict(
            hide_title=True,
            hide_path=True,
            show_breadcrumb=True,
            show_history_navigation=True,
        )
        s.menu_paths.update_menu_path(uuid=app_id, **new_vals)
        app_updated: Dict = s.menu_paths.get_menu_path(uuid=app_id)

        assert app_updated['hideTitle']
        assert app_updated['hidePath']
        assert app_updated['showBreadcrumb']
        assert app_updated['showHistoryNavigation']

        #########
        # Revert the change
        #########
        s.menu_paths.update_menu_path(uuid=app_id, **old_vals)
        app_reverted: Dict = s.menu_paths.get_menu_path(uuid=app_id)

        assert app_reverted['hideTitle'] == old_vals['hide_title']
        assert app_reverted['hidePath'] == old_vals['hide_path']
        assert app_reverted['showBreadcrumb'] == old_vals['show_breadcrumb']
        assert app_reverted['showHistoryNavigation'] == old_vals['show_history_navigation']

    def test_create_and_delete_app(self):
        s.menu_paths.get_menu_path(name='auto-test')
        s.menu_paths.delete_menu_path(name='auto-test')

        app: Dict = s.menu_paths.get_menu_path(name='auto-test')

        assert app

        result: Dict = s.menu_paths.delete_menu_path(name='auto-test')

        assert result

        self.check_app_not_exists(app['id'])

    def test_get_app_reports(self):
        reports: List[Dict] = s.menu_paths.get_menu_path_components(name=app_name)
        assert reports
        assert reports[0]

    def test_get_app_report_ids(self):
        report_ids: List[str] = [r['id'] for r in s.menu_paths.get_menu_path_components(name=app_name)]
        assert report_ids
        assert len(report_ids[0]) > 0
        assert isinstance(report_ids[0], str)

    def test_get_app_path_names(self):
        path_names: List[str] = s.menu_paths.get_menu_path_sub_paths(name=app_name)
        assert path_names

    def test_get_app_activities(self):
        s.set_menu_path(app_name)
        for activity in s.menu_paths.get_menu_path_activities(name=app_name):
            s.activities.delete_activity(uuid=activity['id'])

        for i in range(3):
            s.activities.create_activity(name=f'auto-test-activity {i}')

        activities: List[Dict] = s.menu_paths.get_menu_path_activities(name=app_name)

        assert len(activities) == 3
        for activity in activities:
            assert activity['name'].startswith('auto-test-activity')

        s.menu_paths.delete_all_menu_path_activities(name=app_name)

        activities: List[Dict] = s.menu_paths.get_menu_path_activities(name=app_name)

        assert len(activities) == 0

    def test_rename_app(self):
        new_name: str = 'auto-test'
        s.menu_paths.update_menu_path(name=app_name, new_name=new_name)
        app: Dict = s.menu_paths.get_menu_path(name=new_name)

        assert app['name'] == new_name

        s.menu_paths.update_menu_path(name=new_name, new_name=app_name)
        app: Dict = s.menu_paths.get_menu_path(name=app_name)

        assert app['name'] == app_name
        self.delete_path()

    def delete_path(self):
        if not mock:
            with self.assertRaises(MenuPathError):
                s.menu_paths.delete_menu_path(name=app_name)

        s.pop_out_of_menu_path()
        s.menu_paths.delete_menu_path(name=app_name)



from typing import Dict, Optional, TYPE_CHECKING

import re
from shimoku.api.base_resource import Resource
if TYPE_CHECKING:
    from shimoku.api.resources.app import App

import logging
logger = logging.getLogger(__name__)


class File(Resource):
    """ File resource class """

    _module_logger = logger
    resource_type = 'file'
    plural = 'files'
    alias_field = 'name'
    elastic_supported = True

    def __init__(self, parent: 'App', uuid: Optional[str] = None, alias: Optional[str] = None,
                 db_resource: Optional[Dict] = None):

        params = dict(
            name=alias,
            fileName=re.sub('[^0-9a-zA-Z]+', '-', alias).lower() if alias else None,
            url='',
            contentType='',
            tags=[],
            metadata={},
        )

        super().__init__(parent=parent, uuid=uuid, db_resource=db_resource, params=params,
                         check_params_before_creation=['name'], params_to_serialize=['metadata'])

    async def delete(self):
        """ Delete the report """
        return await self._base_resource.delete()



from typing import Optional, Union, TYPE_CHECKING

from pandas import DataFrame

from shimoku.plt.EChart_definitions.default_echart_options import get_common_echart_options, get_common_series_options

if TYPE_CHECKING:
    from shimoku.plt.plt_layer import PlotLayer


async def funnel_chart(
    self: 'PlotLayer', order: int, names: str, values: str,
    data: Union[str, DataFrame, list[dict]], title: Optional[str] = None,
    rows_size: Optional[int] = None, cols_size: Optional[int] = None,
    padding: Optional[str] = None,
    option_modifications: Optional[dict] = None
):
    """ Create a scatter chart """
    common_options = get_common_echart_options()
    series_options = get_common_series_options()
    del common_options['xAxis']
    del common_options['yAxis']
    del common_options['toolbox']['feature']['magicType']
    common_options['legend']['data'] = '#set_data#'
    common_options['tooltip'] = {
        'trigger': 'item',
        'formatter': "{a} <br/>{b} : {c}%"
    }
    series_options.update({
        'type': 'funnel',
        'left': '10%',
        'top': "10%",
        'bottom': '10%',
        'width': '80%',
        'min': 0,
        'max': 100,
        'minSize': '0%',
        'maxSize': '100%',
        'sort': 'descending',
        'gap': 2,
        'label': {
            'show': True,
            'position': 'inside'
        },
        'labelLine': {
            'length': 10,
            'lineStyle': {
                'width': 1,
                'type': 'solid'
            }
        },
        'itemStyle': {
            'borderColor': '#fff',
            'borderWidth': 1
        },
        'emphasis': {
            'label': {
                'fontSize': 20
            }
        }
    })
    common_options.update({
        'series': series_options
    })

    await self._create_echart(
        fields=[names, {'name': names, 'value': values}], options=common_options,
        data_mapping_to_tuples=await self._choose_data(order, data), option_modifications=option_modifications,
        order=order, title=title, rows_size=rows_size, cols_size=cols_size, padding=padding)



from typing import Optional, Union, TYPE_CHECKING

from pandas import DataFrame

from shimoku.plt.utils import validate_data_is_pandarable, ShimokuPalette
from shimoku.plt.EChart_definitions.default_echart_options import get_common_echart_options, get_common_series_options

if TYPE_CHECKING:
    from shimoku.plt.plt_layer import PlotLayer


async def scatter_chart(
    self: 'PlotLayer', order: int, point_fields: Union[list[tuple[str, str]], tuple[str, str]],
    data: Union[str, DataFrame, list[dict]], x_axis_name: Optional[str] = None, y_axis_name: Optional[str] = None,
    title: Optional[str] = None, rows_size: Optional[int] = None, cols_size: Optional[int] = None,
    padding: Optional[str] = None, show_values: Union[list[str], str, None] = None,
    option_modifications: Optional[dict] = None, variant: Optional[str] = None
):
    """ Create a scatter chart """
    common_options = get_common_echart_options()
    series_options = get_common_series_options()
    x_axis = common_options['xAxis'][0]
    del x_axis['data']
    x_axis['type'] = 'value'
    x_axis['scale'] = True
    common_options['yAxis'][0]['scale'] = True
    del common_options['toolbox']['feature']['magicType']
    series_options['type'] = 'scatter'
    series_options['symbolSize'] = 10
    del series_options['itemStyle']
    del series_options['smooth']
    await self._create_trend_chart(
        echart_options=common_options, series_options=series_options, option_modifications=option_modifications,
        axes=[], values=point_fields, data_mapping_to_tuples=await self._choose_data(order, data),
        order=order, x_axis_names=x_axis_name, y_axis_names=y_axis_name, title=title,
        rows_size=rows_size, cols_size=cols_size, padding=padding, show_values=show_values, variant=variant
    )


async def scatter_with_effect_chart(
        self: 'PlotLayer', order: int, x: str, y: str,
        data: Optional[Union[list[str], DataFrame, list[dict]]],
        x_axis_name: Optional[str] = None, y_axis_name: Optional[str] = None,
        title: Optional[str] = None, rows_size: Optional[int] = None, cols_size: Optional[int] = None,
        padding: Optional[str] = None, effect_points: Optional[list] = None,
        option_modifications: Optional[dict] = None,
):
    """ Create a scatter chart """
    data_mappings_to_tuples = await self._choose_data(order, data)
    if isinstance(data, str):
        data = self._shared_data[data]

    df: DataFrame = validate_data_is_pandarable(data)
    df = df[[x, y]].dropna()
    matrix = df.values

    common_options = get_common_echart_options()
    x_axis = common_options['xAxis'][0]
    del x_axis['data']
    x_axis['type'] = 'value'
    x_axis['scale'] = True
    common_options['yAxis'][0]['scale'] = True
    del common_options['toolbox']['feature']['magicType']
    common_options['series'] = [
        {
            'data': [(point if isinstance(point, list) else list(matrix[point]))
                     for point in effect_points],
            'type': 'effectScatter',
            'symbolSize': 20,
        },
        {
            'type': 'scatter',
            'itemStyle': {
                'color': ShimokuPalette.CHART_C2.value,
            },
            'data': '#set_data#'
        }
    ]

    await self._create_echart(
        options=common_options, fields=[(x, y)], data_mapping_to_tuples=data_mappings_to_tuples,
        order=order, title=title, rows_size=rows_size, cols_size=cols_size,
        padding=padding, option_modifications=option_modifications
    )



from os import getenv
from shimoku.exceptions import CacheError, APIError
import unittest
from utils import initiate_shimoku

s = initiate_shimoku()

business_id: str = getenv('BUSINESS_ID')
environment: str = getenv('ENVIRONMENT')
mock: bool = getenv('MOCK') == 'TRUE'
playground: bool = getenv('PLAYGROUND') == 'TRUE'

s.set_workspace(uuid=business_id)


def delete_dashboard_if_exists(name):
    if s.boards.get_board(name=name):
        s.boards.force_delete_board(name=name)


class TestDashboardMetadataApi(unittest.TestCase):

    def test_dashboard_CRUD(self):
        name = 'Testing dashboard CRUD'
        delete_dashboard_if_exists(name)
        delete_dashboard_if_exists(name + ' updated')

        dashboard = s.boards.create_board(name=name)

        assert dashboard['name'] == name

        dashboard = s.boards.get_board(uuid=dashboard['id'])
        assert dashboard['name'] == name

        s.boards.update_board(uuid=dashboard['id'], new_name=name + ' updated', is_public=True)

        dashboard = s.boards.get_board(uuid=dashboard['id'])
        assert dashboard['name'] == name+' updated'

        assert s.boards.create_board(name=name)
        with self.assertRaises(CacheError):
            s.boards.create_board(name=name)
        with self.assertRaises(CacheError):
            s.boards.create_board(name=name + ' updated')

        s.boards.delete_board(name=name)
        s.boards.delete_board(name=name + ' updated')

        assert not s.boards.get_board(name=name)
        assert not s.boards.get_board(name=name + ' updated')

    def test_create_and_delete_app_dashboard_link(self):
        name = 'Testing dashboard for appdashboards'

        app_name = 'Testing app'
        s.menu_paths.delete_menu_path(uuid=s.menu_paths.get_menu_path(name=app_name)['id'])

        delete_dashboard_if_exists(name)
        delete_dashboard_if_exists(name + ' updated')

        s.boards.create_board(name=name)
        app_id = s.menu_paths.get_menu_path(name=app_name)['id']

        s.boards.add_menu_path_in_board(name=name, menu_path_id=app_id)

        app_ids = s.boards.get_board_menu_path_ids(name=name)
        assert app_id == app_ids[0]

        if not mock and not playground:
            with self.assertRaises(APIError):
                s.boards.delete_board(name=name)

        s.boards.remove_menu_path_from_board(name=name, menu_path_id=app_id)
        s.boards.delete_board(name=name)
        s.menu_paths.delete_menu_path(uuid=app_id)

    def test_delete_all_app_dashboards_links(self):
        name = 'Testing dashboard for appdashboards delete all'

        app_names = ['Testing app 1', 'Testing app 2', 'Testing app 3']

        for app_name in app_names:
            s.menu_paths.delete_menu_path(uuid=s.menu_paths.get_menu_path(name=app_name)['id'])

        delete_dashboard_if_exists(name)

        app_ids = []
        s.boards.create_board(name=name)
        for app_name in app_names:
            app_id = s.menu_paths.get_menu_path(name=app_name)['id']
            s.boards.add_menu_path_in_board(name=name, menu_path_id=app_id)
            app_ids.append(app_id)

        dashboard_app_ids = s.boards.get_board_menu_path_ids(name=name)
        assert len(dashboard_app_ids) == len(app_names)

        s.boards.remove_all_menu_paths_from_board(name=name)

        dashboard_app_ids = s.boards.get_board_menu_path_ids(name=name)
        assert len(dashboard_app_ids) == 0

        for app_id in app_ids:
            s.boards.add_menu_path_in_board(name=name, menu_path_id=app_id)

        s.boards.force_delete_board(name=name)

        assert not s.boards.get_board(name=name)

        for app_id in app_ids:
            s.menu_paths.delete_menu_path(uuid=app_id)

    def test_group_apps(self):
        name = 'Testing dashboard for grouping apps'

        app_names = ['Testing app 1', 'Testing app 2', 'Testing app 3']

        for app_name in app_names:
            s.menu_paths.delete_menu_path(uuid=s.menu_paths.get_menu_path(name=app_name)['id'])

        delete_dashboard_if_exists(name)

        app_ids = []
        s.boards.create_board(name=name)
        for app_name in app_names:
            app_id = s.menu_paths.get_menu_path(name=app_name)['id']
            app_ids.append(app_id)

        s.boards.group_menu_paths(menu_path_names=app_names, name=name)
        dashboard_app_ids = s.boards.get_board_menu_path_ids(name=name)

        for app_id, app_name in zip(app_ids, app_names):
            assert app_id in dashboard_app_ids
            s.boards.remove_menu_path_from_board(menu_path_name=app_name, name=name)
            dashboard_app_ids.remove(app_id)

        assert len(dashboard_app_ids) == 0

        s.boards.group_menu_paths(menu_path_ids=app_ids, name=name)
        dashboard_app_ids = s.boards.get_board_menu_path_ids(name=name)

        for app_id, app_name in zip(app_ids, app_names):
            assert app_id in dashboard_app_ids
            s.boards.remove_menu_path_from_board(menu_path_name=app_name, name=name)
            dashboard_app_ids.remove(app_id)

        assert len(dashboard_app_ids) == 0

        all_app_ids = s.workspaces.get_workspace_menu_path_ids(uuid=business_id)

        s.boards.group_menu_paths(menu_path_names='all', name=name)
        dashboard_app_ids = s.boards.get_board_menu_path_ids(name=name)

        for app_id in all_app_ids:
            assert app_id in dashboard_app_ids
            s.boards.remove_menu_path_from_board(menu_path_id=app_id, name=name)
            dashboard_app_ids.remove(app_id)

        assert len(dashboard_app_ids) == 0

        s.boards.delete_board(name=name)

        assert not s.boards.get_board(name=name)

        for app_id in app_ids:
            s.menu_paths.delete_menu_path(uuid=app_id)






from typing import Optional, TYPE_CHECKING

from pandas import DataFrame

from shimoku.plt.EChart_definitions.default_echart_options import get_common_echart_options, get_common_series_options

if TYPE_CHECKING:
    from shimoku.plt.plt_layer import PlotLayer


async def sankey_chart(
    self: 'PlotLayer', order: int, sources: str,
    targets: str, values: str, data: list[dict],
    option_modifications: Optional[dict] = None,
    title: Optional[str] = None, padding: Optional[str] = None,
    rows_size: Optional[int] = None, cols_size: Optional[int] = None
):
    """ Create a sankey chart """
    df = DataFrame(data)
    common_options = get_common_echart_options()
    series_options = get_common_series_options()
    del common_options['xAxis']
    del common_options['yAxis']
    del common_options['toolbox']['feature']['magicType']
    del common_options['grid']
    names = df[sources].to_list() + df[targets].to_list()
    series_options.update({
        'type': 'sankey',
        'layout': 'none',
        'emphasis': {
            'focus': 'adjacency',
        },
        'data': [{'name': name} for i, name in enumerate(names) if name not in names[:i]],
        'links': '#set_data#',
    })
    common_options['series'] = [series_options]
    await self._create_echart(
        data_mapping_to_tuples=await self._choose_data(order, data),
        fields=[{'source': sources, 'target': targets, 'value': values}],
        options=common_options, order=order, title=title, option_modifications=option_modifications,
        rows_size=rows_size, cols_size=cols_size, padding=padding)



from typing import Optional, Union

import asyncio

from uuid import uuid1

from shimoku.api.resources.role import user_delete_role, user_get_role, user_get_roles, user_create_role
from shimoku.api.resources.dashboard import Dashboard
from shimoku.api.resources.business import Business

import logging
from shimoku.execution_logger import ClassWithLogging
logger = logging.getLogger(__name__)


class BoardsLayer(ClassWithLogging):
    """
    This class is used to interact with the API at the board level.
    """
    
    _module_logger = logger
    _use_info_logging = True

    def __init__(
            self, business: Business,
    ):
        self._business = business
        self._get_for_roles = self._business.get_app if business else None
    
    async def _get_dashboard_with_warning(
        self, uuid: Optional[str] = None, name: Optional[str] = None
    ) -> Optional[Dashboard]:
        """
        Get the dashboard metadata
        :param name: name of the dashboard
        :param uuid: UUID of the dashboard
        :return: The dashboard object
        """
        dashboard: Dashboard = await self._business.get_dashboard(uuid=uuid, name=name, create_if_not_exists=False)
        if not dashboard:
            logger.warning(f"Board with name {name} not found, not doing anything")
        return dashboard
    
    async def create_board(
        self, name: str, order: Optional[int] = None,
        is_public: bool = False, is_disabled: bool = False
    ) -> dict:
        """
        Create a board
        :param name: name of the board
        :param order: order of the board
        :param is_public: is the board public
        :param is_disabled: is the board disabled
        :return: board metadata
        """
        return (await self._business.create_dashboard(
            name=name, order=order if order is not None else len(await self._business.get_dashboards()),
            is_disabled=is_disabled, is_public=is_public
        )).cascade_to_dict()

    async def get_board(
        self, uuid: Optional[str] = None, name: Optional[str] = None
    ) -> Optional[dict]:
        """
        Get the board metadata
        :param name: name of the board
        :param uuid: UUID of the board
        :return: board metadata
        """
        dashboard = await self._get_dashboard_with_warning(uuid=uuid, name=name)
        if not dashboard:
            return None
        return dashboard.cascade_to_dict()
    
    async def delete_board(
        self, uuid: Optional[str] = None, name: Optional[str] = None
    ) -> bool:
        """
        Delete a board
        :param name: name of the board
        :param uuid: UUID of the board
        """
        return await self._business.delete_dashboard(uuid=uuid, name=name)
    
    async def update_board(
        self, uuid: Optional[str] = None, name: Optional[str] = None,
        new_name: Optional[str] = None, order: Optional[int] = None,
        is_public: Optional[bool] = None, is_disabled: Optional[bool] = None
    ) -> bool:
        """
        Update a board
        :param new_name: name of the board
        :param uuid: UUID of the board
        :param name: new name of the board
        :param order: new order of the board
        :param is_public: new public permission of the board
        :param is_disabled: new is_disabled of the board
        """
        return await self._business.update_dashboard(
            uuid=uuid, name=name, new_name=new_name, order=order,
            isDisabled=is_disabled,
            publicPermission={
                'isPublic': is_public,
                'permission': 'READ',
                'token': str(uuid1())
            } if is_public is not None else None
        )
    
    async def is_menu_path_in_board(
        self, menu_path_name: Optional[str] = None, menu_path_id: Optional[str] = None,
        uuid: Optional[str] = None, name: Optional[str] = None
    ) -> bool:
        """
        Check if an app is in a board
        :param menu_path_name: name of the board
        :param menu_path_id: UUID of the board
        :param name: name of the board
        :param uuid: UUID of the board
        :return: True if the app is in the board, False otherwise
        """
        dashboard = await self._get_dashboard_with_warning(uuid=uuid, name=name)
        if not dashboard:
            return False
        app = await self._business.get_app(name=menu_path_name, uuid=menu_path_id, create_if_not_exists=False)
        if not app:
            return False
        return menu_path_id in await dashboard.list_app_ids()
    
    async def add_menu_path_in_board(
        self, menu_path_name: Optional[str] = None, menu_path_id: Optional[str] = None,
        uuid: Optional[str] = None, name: Optional[str] = None
    ):
        """
        Add an app in a board
        :param menu_path_name: name of the board
        :param menu_path_id: UUID of the board
        :param name: name of the board
        :param uuid: UUID of the board
        """
        dashboard = await self._business.get_dashboard(uuid=uuid, name=name)
        app = await self._business.get_app(name=menu_path_name, uuid=menu_path_id)
        await dashboard.insert_app(app)

    async def remove_menu_path_from_board(
        self, menu_path_name: Optional[str] = None, menu_path_id: Optional[str] = None,
        uuid: Optional[str] = None, name: Optional[str] = None
    ):
        """
        Remove an app from a board
        :param menu_path_name: name of the board
        :param menu_path_id: UUID of the board
        :param name: name of the board
        :param uuid: UUID of the board
        """
        dashboard: Dashboard = await self._get_dashboard_with_warning(uuid=uuid, name=name)
        if dashboard:
            app = await self._business.get_app(name=menu_path_name, uuid=menu_path_id, create_if_not_exists=False)
            if not app:
                logger.warning(f"Menu path {menu_path_name if menu_path_name else menu_path_id} "
                               f"not found, not doing anything")
                return

            await dashboard.remove_app(app)
        
    async def remove_all_menu_paths_from_board(
        self, uuid: Optional[str] = None, name: Optional[str] = None
    ):
        """
        Delete all boards links of a board
        :param name: name of the board
        :param uuid: UUID of the board
        """
        dashboard: Dashboard = await self._get_dashboard_with_warning(uuid=uuid, name=name)
        if dashboard:
            await dashboard.remove_all_apps()
        
    async def group_menu_paths(
        self, menu_path_names: Union[Optional[list[str]], str] = None,
        menu_path_ids: Optional[list[str]] = None,
        uuid: Optional[str] = None, name: Optional[str] = None
    ):
        """
        Add multiple boards in a board, if the board does not exist create it
        :param menu_path_names: list of board names
        :param menu_path_ids: list of board UUIDs
        :param name: name of the board
        :param uuid: UUID of the board
        """
        dashboard = await self._business.get_dashboard(uuid=uuid, name=name)

        apps = []
        if menu_path_names:
            if isinstance(menu_path_names, list):
                apps = await asyncio.gather(*[self._business.get_app(name=menu_path) for menu_path in menu_path_names])
            elif menu_path_names == 'all':
                apps = await self._business.get_apps()
        elif menu_path_ids:
            apps = await asyncio.gather(*[self._business.get_app(uuid=app_id) for app_id in menu_path_ids])

        if not apps:
            logger.warning('No boards have been provided, not doing anything')
            return

        await asyncio.gather(*[dashboard.insert_app(app=app) for app in apps])
    
    async def get_board_menu_path_ids(
        self, uuid: Optional[str] = None, name: Optional[str] = None
    ) -> list[str]:
        """
        Get the list of the ids of the boards in the board
        :param name: name of the board
        :param uuid: UUID of the board
        :return: list of the ids of the boards in the board
        """
        dashboard: Dashboard = await self._get_dashboard_with_warning(uuid=uuid, name=name)
        if dashboard:
            return await dashboard.list_app_ids()
    
    async def force_delete_board(
        self, uuid: Optional[str] = None, name: Optional[str] = None
    ):
        """
        Force delete a board, this will delete the links between the board and the boards first, so that
        the board can always be deleted, but the apps will not be deleted
        :param name: name of the board
        :param uuid: UUID of the board
        """
        dashboard: Dashboard = await self._get_dashboard_with_warning(uuid=uuid, name=name)
        if dashboard:
            await dashboard.remove_all_apps()
            await self._business.delete_dashboard(uuid=uuid, name=name)
        
    # Role management
    get_role = user_get_role
    get_roles = user_get_roles
    create_role = user_create_role
    delete_role = user_delete_role



""""""
from os import getenv
import inspect

import shimoku_api_python as shimoku


api_key: str = getenv('API_TOKEN')
universe_id: str = getenv('UNIVERSE_ID')
business_id: str = getenv('BUSINESS_ID')
environment: str = getenv('ENVIRONMENT')
verbosity: str = getenv('VERBOSITY')


s = shimoku.Client(
    access_token=api_key,
    universe_id=universe_id,
    environment=environment,
    verbosity=verbosity
)
s.set_workspace(uuid=business_id)

s.activate_async_execution()

components = [f for _, f in s.html_components.__dict__.items() if callable(f)]

parameters = dict(
    background_url="https://images.unsplash.com/photo-1553356084-58ef4a67b2a7?ixlib=rb-4.0.3&ixid=MnwxMjA3fDB8MHxwaG90by1wYWdlfHx8fGVufDB8fHx8&auto=format&fit=crop&w=774&q=80",
    background="https://images.unsplash.com/photo-1553356084-58ef4a67b2a7?ixlib=rb-4.0.3&ixid=MnwxMjA3fDB8MHxwaG90by1wYWdlfHx8fGVufDB8fHx8&auto=format&fit=crop&w=774&q=80",
    image="https://images.unsplash.com/photo-1553356084-58ef4a67b2a7?ixlib=rb-4.0.3&ixid=MnwxMjA3fDB8MHxwaG90by1wYWdlfHx8fGVufDB8fHx8&auto=format&fit=crop&w=774&q=80",
    image_option='animated',
    title="Lorem ipsum",
    subtitle="Lorem ipsum dolor sit amet",
    href="https://shimoku.com",
    text='Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore ' \
           'et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut ' \
           'aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum ' \
           'dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia ' \
           'deserunt mollit anim id est laborum.',
    main_text='Lorem ipsum',
    internal_text='Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore ' \
              'et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut ' \
              'aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum ' \
              'dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia ' \
              'deserunt mollit anim id est laborum.',
    note='Lorem ipsum',
    button_panel='Lorem ipsum',
    button_text="Lorem ipsum",
    symbol_name='insights',
    h1="Lorem ipsum",
    line="Lorem ipsum",
    modal_title="Lorem ipsum",
    background_color='var(--color-primary)',
)

for component_function in components:
    function_signature = inspect.signature(component_function)
    call_parameters = {parameter[0]: parameter[1]
                       for parameter in parameters.items() if parameter[0] in function_signature.parameters}
    print(component_function)
    print(function_signature.parameters)
    #If the parameters dont exist in the function signature, skip it
    if not call_parameters:
        continue
    print(f'passed {component_function.__name__}')
    html = component_function(**call_parameters)
    s.set_menu_path(f"Components test", f"{component_function.__name__}")
    s.plt.html(html=html, order=0)

s.run()
